__author__ = 'Joe Linn'

from .abstract import AbstractSuggestion


class Term(AbstractSuggestion):
    """
    @see: http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/search-suggesters-term.html
    """

    SORT_SCORE = 'score'
    SORT_FREQUENCY = 'frequency'

    SUGGEST_MODE_MISSING = 'missing'
    SUGGEST_MODE_POPULAR = 'popular'
    SUGGEST_MODE_ALWAYS = 'always'

    def set_analyzer(self, analyzer):
        """
        Set the analyzer to be used for this suggest operation
        @param analyzer: the name of a pre-defined analyzer
        @type analyzer: str
        @return:
        @rtype: self
        """
        return self.set_param("analyzer", analyzer)

    def set_sort(self, sort):
        """
        Set the sorting method for results
        @param sort: see SORT_* constants for options
        @type sort: str
        @return:
        @rtype: self
        """
        return self.set_param("sort", sort)

    def set_suggest_mode(self, mode):
        """

        @param mode: see SUGGEST_MODE_* constants for options
        @type mode: str
        @return:
        @rtype: self
        """
        return self.set_param("suggest_mode", mode)

    def set_lowercase_terms(self, lowercase=True):
        """
        If true, suggest terms will be lower cased after text analysis
        @param lowercase:
        @type lowercase: bool
        @return:
        @rtype: self
        """
        return self.set_param("lowercase_terms", bool(lowercase))

    def set_max_edits(self, max_edits):
        """
        Set the maximum edit distance candidate suggestions can have in order to be considered as a suggestion
        @param max_edits: A value between 1 and 2. Defaults to 2.
        @type max_edits: float
        @return:
        @rtype: self
        """
        return self.set_param("max_edits", max_edits)

    def set_prefix_length(self, length):
        """
        The number of minimum prefix characters that must match in order to be a suggestion candidate
        @param length: Defaults to 1
        @type length: int
        @return:
        @rtype: self
        """
        return self.set_param("prefix_len", length)

    def set_min_word_length(self, length):
        """
        The minimum length a suggest text term must have in order to be included.
        @param length: Defaults to 4
        @type length: int
        @return:
        @rtype: self
        """
        return self.set_param("min_word_len", length)

    def set_max_inspections(self, max_inspections):
        """

        @param max_inspections: Defaults to 5
        @type max_inspections: int
        @return:
        @rtype: self
        """
        return self.set_param("max_inspections", max_inspections)

    def set_min_doc_frequency(self, freq):
        """
        Set the minimum number of documents in which a suggestion should appear
        @param freq: Defaults to 0. If the value is greater than 1, it must be a whole number.
        @type freq: int or float
        @return:
        @rtype: self
        """
        return self.set_param("min_doc_freq", freq)

    def set_max_term_frequency(self, freq):
        """
        Set the maximum number of documents in which a suggest text token can exist in order to be included
        @param freq:
        @type freq: float or int
        @return:
        @rtype: self
        """
        return self.set_param("max_term_freq", freq)


__author__ = 'Joe Linn'

import pylastica.connection
import pylastica.index
import pylastica.request


class Client(object):

    def __init__(self, host=None, port=None, path=None, url=None, transport=None, persistent=True, timeout=None, connections=None, round_robin=False, log=False, retry_on_conflict=0, callback=None):
        """
        @param host:
        @type host: str
        @param port:
        @type port: int
        @param path:
        @type path: str
        @param url:
        @type url: str
        @param transport:
        @type transport:
        @param persistent:
        @type persistent:
        @param timeout:
        @type timeout:
        @param connections: A list of connection parameters [{'host': 'localhost', 'port': 9200}]
        @type connections: list of dict
        @param round_robin:
        @type round_robin:
        @param log: set to True to enable logging. Set to a string to specify a log file.
        @type log: bool or str
        @param retry_on_conflict: retry document update on conflict
        @type retry_on_conflict: bool
        @param callback:
        @type callback:
        @return:
        @rtype:
        """
        if connections is None:
            connections = []
        self._connections = []
        self._callback = callback
        """@type: pylastica.response.Response"""
        self._last_response = None
        self._config = {
            'host': host,
            'port': port,
            'path': path,
            'url': url,
            'transport': transport,
            'persistent': persistent,
            'timeout': timeout,
            'connections': connections,
            'round_robin': round_robin,
            'log': log,
            'retry_on_conflict': retry_on_conflict
        }
        self._init_connections()

    def _init_connections(self):
        connections = self.get_config('connections')
        for connection in connections:
            self._connections.append(pylastica.connection.Connection.create(connection))
        if 'servers' in self._config:
            for server in self.get_config('servers'):
                self._connections.append(pylastica.connection.Connection.create(server))
        if not len(self._connections):
            self._connections.append(pylastica.connection.Connection.create(self._configure_params()))

    def _configure_params(self):
        """

        @return:
        @rtype: dict
        """
        config = self.get_config()
        params = {
            'config': {}
        }
        for key, value in config.iteritems():
            if key in ['curl', 'headers', 'url']:
                params['config'][key] = value
            else:
                params[key] = value
        return params

    def set_config(self, config):
        """
        Set specific config values (updates if values are already set)
        @param config:
        @type config: dict
        @return:
        @rtype: self
        """
        for key, value in config.iteritems():
            self._config[key] = value
        return self

    def get_config(self, key=None):
        """
        Return a specific config value of the whole config dict
        @param key:
        @type key: str
        @return:
        @rtype: str or dict
        """
        if key is None:
            return self._config
        if key not in self._config:
            raise pylastica.exception.InvalidException("Config key is not set: %s" % key)
        return self._config[key]

    def set_config_value(self, key, value):
        """
        Set / overwrite a specific config value
        @param key:
        @type key: str
        @param value:
        @type value: mixed
        @return:
        @rtype: self
        """
        return self.set_config({key: value})

    def get_config_value(self, keys, default=None):
        """

        @param keys: a string to get a single config value, or a list of keys to get multiple config values
        @type keys: list or str
        @param default: default value which will be returned if the key is not found in the config
        @type default: mixed
        @return:
        @rtype: mixed
        """
        value = self._config
        if isinstance(keys, str):
            keys = [keys]
        for key in keys:
            if key in value:
                value = value[key]
            else:
                return default
        return value

    def get_index(self, name):
        """
        Returns an index object for the given index name
        @param name: index name
        @type name: str
        @return:
        @rtype: pylastica.index.Index
        """
        return pylastica.index.Index(self, name)

    def add_header(self, header, header_value):
        """
        Add a HTTP header
        @param header: header key
        @type header: str
        @param header_value: header value
        @type header_value: str
        @return:
        @rtype: self
        """
        if isinstance(header, str) and isinstance(header_value, str):
            self._config['headers'][header] = header_value
            return self
        raise pylastica.exception.InvalidException("header and header_value must be strings.")

    def remove_header(self, header):
        """
        Remove a HTTP header
        @param header: key of header to remove
        @type header: str
        @return:
        @rtype: self
        """
        if isinstance(header, str):
            if header in self._config['headers']:
                del self._config['headers'][header]
            return self
        else:
            raise pylastica.exception.InvalidException("header must be of type str: %r" %header)

    def add_documents(self, docs):
        """
        Uses _bulk to send documents to the server
        @param docs: list of document objects
        @type docs: list of pylastica.document.Document
        @return: bulk response object
        @rtype: pylastica.bulk.responseset.ResponseSet
        """
        assert isinstance(docs, list) and len(docs), "docs must be a list of at least one Document object: %r" % docs
        return pylastica.bulk.Bulk(self).add_documents(docs).send()

    def update_document(self, doc_id, data, index, doc_type, options=None):
        """
        Update a document via an update script
        @param doc_id:  document id
        @type doc_id: str
        @param data: data for request body
        @type data: dict or pylastica.script.Script or pylastica.document.Document
        @param index: index to update
        @type index: str
        @param doc_type: document type
        @type doc_type: str
        @param options: optional dict of query parameters
        @type options: dict
        @return:
        @rtype: pylastica.response.Response
        """
        if options is None:
            options = {}
        path = '/'.join([index, doc_type, str(doc_id), '_update'])
        if isinstance(data, pylastica.script.Script):
            request_data = data.to_dict()
        elif isinstance(data, pylastica.document.Document):
            request_data = {'doc': data.data}
            if data.doc_as_upsert:
                request_data['doc_as_upsert'] = True
            doc_options = data.get_options([
                'version',
                'version_type',
                'routing',
                'percolate',
                'parent',
                'fields',
                'retry_on_conflict',
                'consistency',
                'replication',
                'refresh',
                'timeout',
            ])
            options.update(doc_options)
            #set fields param to source only if options was not set before
            if data.auto_populate and (data.is_auto_populate() or self.get_config_value(['document', 'autopopulate'], False)) and 'fields' not in options:
                options['fields'] = '_source'
        else:
            request_data = data
        if isinstance(data, pylastica.script.Script) or isinstance(data, pylastica.document.Document):
            if data.has_upsert():
                request_data['upsert'] = data.upsert.data
        if 'retry_on_conflict' not in options:
            retry_on_conflict = self.get_config('retry_on_conflict')
            options['retry_on_conflict'] = retry_on_conflict
        response = self.request(path, pylastica.request.Request.POST, request_data, options)
        if response.is_ok() and isinstance(data, pylastica.document.Document) and data.auto_populate or self.get_config_value(['document', 'autopopulate'], False):
            response_data = response.data
            if '_version' in response_data:
                data.version = response_data['_version']
            if 'fields' in options:
                self._populate_document_fields_from_response(response, data, options['fields'])
        return response

    def _populate_document_fields_from_response(self, response, document, fields):
        """

        @param response:
        @type response: pylastica.response.Response
        @param document:
        @type document: pylastica.document.Document
        @param fields: list of field names to be populated, or '_source' if whole document should be updated
        @type fields: list of str or str
        """
        response_data = response.data
        if fields == '_source':
            if '_source' in response_data['get'] and isinstance(response['get']['_source'], dict):
                document.data = response['get']['_source']
        else:
            keys = fields.split(',')
            data = document.data
            for key in keys:
                if key in response_data['get']['fields']:
                    data[key] = response_data['get']['fields'][key]
                elif key in data:
                    del data[key]
            document.data = data

    def delete_documents(self, docs):
        """
        Bulk delete documents
        @param docs:
        @type docs: list of pylastica.document.Document
        @return:
        @rtype: pylastica.bulk.ResponseSet
        """
        assert isinstance(docs, list) and len(docs) > 0, "docs must be a list of at least one document: %r" % docs
        from pylastica.bulk.action import Action
        return pylastica.bulk.Bulk(self).add_documents(docs, Action.OP_TYPE_DELETE).send()

    @property
    def status(self):
        """
        Returns the status object for all indices
        @return:
        @rtype: pylastica.status.Status
        """
        return pylastica.status.Status(self)

    @property
    def cluster(self):
        """
        Returns the current cluster
        @return:
        @rtype: pylastica.cluster.Cluster
        """
        return pylastica.cluster.Cluster(self)

    def add_connection(self, connection):
        """

        @param connection:
        @type connection: pylastica.connection.Connection
        @return:
        @rtype: self
        """
        self._connections.append(connection)
        return self

    def get_connection(self):
        """

        @return:
        @rtype: pylastica.connection.Connection
        """
        enabled_connection = None
        for connection in self._connections:
            if connection.is_enabled():
                enabled_connection = connection
                break
        if enabled_connection is None:
            raise pylastica.exception.ClientException("No enabled connection.")
        return enabled_connection

    @property
    def connections(self):
        """

        @return:
        @rtype: list of pylastica.connection.Connection
        """
        return self._connections

    @connections.setter
    def connections(self, connections):
        """

        @param connections:
        @type connections: list of pylastica.connection.Connection
        """
        self._connections = connections

    def delete_ids(self, ids, index, doc_type):
        """
        Deletes documents with the given ids, index, and doc_type
        @param ids: document ids
        @type ids: list of str
        @param index: index / alias name
        @type index: str or pylastica.index.Index
        @param doc_type: type of documents
        @type doc_type: str
        @return:
        @rtype: pylastica.bulk.ResponseSet
        """
        assert isinstance(ids, list) and len(ids) > 0, "ids must be a list of str: %r" % ids
        bulk = pylastica.bulk.Bulk(self)
        bulk.index = index
        bulk.doc_type = doc_type
        for doc_id in ids:
            action = pylastica.bulk.action.Action(pylastica.bulk.action.Action.OP_TYPE_DELETE)
            action.set_id(doc_id)
            bulk.add_action(action)
        return bulk.send()

    def bulk(self, params):
        """
        Bulk operation
        @see: http://www.elasticsearch.org/guide/reference/api/bulk.html
        @param params: parameter list
        @type params: list of dict
        @return:
        @rtype: pylastica.bulk.ResponseSet
        """
        assert isinstance(params, list) and len(params) > 0, "params must be a list: %r" %params
        bulk = pylastica.bulk.Bulk(self)
        return bulk.add_raw_data(params).send()

    def request(self, path, method=pylastica.request.Request.GET, data=None, query=None):
        """
        Makes a call to the elasticsearch server.
        @param path: path to call
        @type path: str
        @param method: REST method to use. See Request class properties.
        @type method: str
        @param data: optional data as dict
        @type data: dict
        @param query: optional query params
        @type query: dict
        @return:
        @rtype: pylastica.response.Response
        """
        connection = self.get_connection()
        try:
            request = pylastica.request.Request(path, method, data, query, connection)
            self._log(request)
            response = request.send()
            self._last_request = request
            self._last_response = response
            return response
        except pylastica.exception.ConnectionException as e:
            connection.enabled = False
            #calls callback with connection as param to make it possible to persist invalid connections
            if self._callback is not None:
                self._callback(connection, e)
            return self.request(path, method, data, query)

    def optimize_all(self, args=None):
        """
        Optimizes all search indices
        @param args: optional arguments
        @type args: dict
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request('_optimize', pylastica.request.Request.POST, args)

    def _log(self, message):
        """

        @param message:
        @type message: str or pylastica.request.Request
        """
        if self.get_config('log'):
            log = pylastica.log.Log(self.get_config('log'))
            log.log(message)

    def get_last_request(self):
        """

        @return:
        @rtype: pylastica.request.Request
        """
        return self._last_request

    def get_last_response(self):
        """

        @return:
        @rtype: pylastica.response.Response
        """
        return self._last_response




__author__ = 'Joe Linn'

import pylastica

class TopChildren(pylastica.query.AbstractQuery):
    def __init__(self, query, doc_type=None):
        """

        @param query: query string or query object
        @type query: str or pylastica.query.Query or pylastica.query.AbstractQuery
        @param doc_type: parent document type
        @type doc_type: str
        """
        self.set_query(query)
        self.set_type(doc_type)

    def set_query(self, query):
        """
        Set the query
        @param query:
        @type query: str or pylastica.query.Query or pylastica.query.AbstractQuery
        @return:
        @rtype: self
        """
        query = pylastica.query.Query.create(query)
        return self.set_param('query', query.to_dict()['query'])

    def set_type(self, doc_type):
        """
        Set the parent document type
        @param doc_type:
        @type doc_type: str
        @return:
        @rtype: self
        """
        return self.set_param('type', doc_type)



__author__ = 'Joe Linn'

import abc
import pylastica.response


class AbstractException(Exception):
    __metaclass__ = abc.ABCMeta


class BulkException(AbstractException):
    pass


class ClientException(AbstractException):
    pass


class ConnectionException(AbstractException):
    def __init__(self, message, request=None, response=None, *args, **kwargs):
        """

        @param message:
        @type message: str
        @param request:
        @type request: pylastica.request.Request
        @param response:
        @type response: pylastica.response.Response
        @param args:
        @type args:
        @param kwargs:
        @type kwargs:
        """
        if request is not None:
            assert isinstance(request, pylastica.Request), "request must be an instance of Request: %r" % request
        if response is not None:
            assert isinstance(response, pylastica.response.Response), "response must be an instance of Response: %r" % response
        self._request = request
        self._response = response
        super(ConnectionException, self).__init__(message, *args, **kwargs)

    @property
    def request(self):
        """

        @return:
        @rtype: pylastica.request.Request
        """
        return self._request

    @property
    def response(self):
        """

        @return:
        @rtype: pylastica.response.Response
        """
        return self._response


class InvalidException(AbstractException):
    pass


class NotFoundException(AbstractException):
    pass


class NotImplementedException(AbstractException):
    pass


class ResponseException(AbstractException):
    def __init__(self, request, response, *args, **kwargs):
        """

        @param request:
        @type request: pylastica.request.Request
        @param response:
        @type response: pylastica.response.Response
        """
        self._request = request
        self._response = response
        super(ResponseException, self).__init__(response.get_error(), *args, **kwargs)

    @property
    def request(self):
        """

        @return:
        @rtype: pylastica.request.Request
        """
        return self._request

    @property
    def response(self):
        """

        @return:
        @rtype: pylastica.response.Response
        """
        return self._response


class RuntimeException(AbstractException):
    pass


__author__ = 'Joe Linn'

from .abstract import AbstractSuggestion


class Phrase(AbstractSuggestion):
    """
    @see: http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/search-suggesters-phrase.html
    """

    def set_analyzer(self, analyzer):
        """
        Set the analyzer for this suggestion
        @param analyzer: a pre-defined Elasticsearch analyzer
        @type analyzer: str
        @return:
        @rtype: self
        """
        return self.set_param("analyzer", analyzer)

    def set_gram_size(self, size):
        """
        Set the max size of the n-grams (shingles) in the field
        @param size:
        @type size: int
        @return:
        @rtype: self
        """
        return self.set_param("gram_size", size)

    def set_real_world_error_likelihood(self, likelihood):
        """
        Set the likelihood of a term being misspelled even if the term exists in the dictionary
        @param likelihood: Defaults to 0.95, meaning 5% of the words are misspelled.
        @type likelihood: float
        @return:
        @rtype: self
        """
        return self.set_param("real_world_error_likelihood", likelihood)

    def set_confidence(self, confidence):
        """
        Set the factor applied to the input phrases score to be used as a threshold for other suggestion candidates.
        Only candidates which score higher than this threshold will be included in the result.
        @param confidence: Defaults to 1.0
        @type confidence: float
        @return:
        @rtype: self
        """
        return self.set_param("confidence", confidence)

    def set_max_errors(self, max_errors):
        """
        Set the maximum percentage of the terms considered to be misspellings in order to form a correction
        @param max_errors:
        @type max_errors: float
        @return:
        @rtype: self
        """
        return self.set_param("max_errors", max_errors)

    def set_separator(self, separator):
        """

        @param separator:
        @type separator: str
        @return:
        @rtype: self
        """
        return self.set_param("separator", separator)

    def set_highlight(self, pre_tag, post_tag):
        """
        Set suggestion highlighting
        @param pre_tag:
        @type pre_tag: str
        @param post_tag:
        @type post_tag: str
        @return:
        @rtype: self
        """
        return self.set_param("highlight", {
            'pre_tag': pre_tag,
            'post_tag': post_tag
        })

    def set_stupid_backoff_smoothing(self, discount = 0.4):
        """

        @param discount:
        @type discount: float
        @return:
        @rtype: self
        """
        return self.set_smoothing_model("stupid_backoff", {"discount": discount})

    def set_laplace_smoothing(self, alpha=0.5):
        """

        @param alpha:
        @type alpha: float
        @return:
        @rtype: self
        """
        return self.set_smoothing_model("laplace", {"alpha": alpha})

    def set_linear_interpolation_smoothing(self, trigram_lambda, bigram_lambda, unigram_lambda):
        """

        @param trigram_lambda:
        @type trigram_lambda: float
        @param bigram_lambda:
        @type bigram_lambda: float
        @param unigram_lambda:
        @type unigram_lambda: float
        @return:
        @rtype: self
        """
        return self.set_smoothing_model("linear_interpolation", {
            "trigram_lambda": trigram_lambda,
            "bigram_lambda": bigram_lambda,
            "unigram_lambda": unigram_lambda
        })

    def set_smoothing_model(self, model, params):
        """

        @param model: the name of the smoothing model
        @type model: str
        @param params:
        @type params: dict
        @return:
        @rtype: self
        """
        return self.set_param("smoothing", {model: params})

    def add_candidate_generator(self, generator):
        """

        @param generator:
        @type generator: pylastica.suggest.candidategenerator.abstract.AbstractCandidateGenerator
        @return:
        @rtype: self
        """
        generator = generator.to_dict()
        return self.add_param(generator.keys()[0], generator.values()[0])


__author__ = 'Joe Linn'

from .shard import *

class Index(object):
    def __init__(self, name, data):
        """

        @param name: name of the index
        @type name: str
        @param data: index health data
        @type data: dict
        """
        super(Index, self).__init__()
        self._name = name
        self._data = data

    @property
    def name(self):
        """
        @return: the name of the index
        @rtype: str
        """
        return self._name

    @property
    def status(self):
        """
        Get the status of the index
        @return: green, yellow, or read
        @rtype: str
        """
        return self._data['status']

    @property
    def number_of_shards(self):
        """
        Get the number of shards in the index
        @return:
        @rtype: int
        """
        return int(self._data['number_of_shards'])

    @property
    def number_of_replicas(self):
        """
        Get the number of replicas
        @return:
        @rtype: int
        """
        return int(self._data['number_of_replicas'])

    @property
    def active_primary_shards(self):
        """
        Get the number of active primary shards
        @return:
        @rtype: int
        """
        return int(self._data['active_primary_shards'])

    @property
    def active_shards(self):
        """
        Get the number of active shards
        @return:
        @rtype: int
        """
        return int(self._data['active_shards'])

    @property
    def relocating_shards(self):
        """
        Get the number of relocating shards
        @return:
        @rtype: int
        """
        return int(self._data['relocating_shards'])

    @property
    def initializing_shards(self):
        """
        Get the number of initializing shards
        @return:
        @rtype: int
        """
        return int(self._data['initializing_shards'])

    @property
    def unassigned_shards(self):
        """
        Get the number of unassigned shards
        @return:
        @rtype: int
        """
        return int(self._data['unassigned_shards'])

    @property
    def shards(self):
        """
        Get the health of the shards in this index
        @return:
        @rtype: list of pylastica.cluster.health.Shard
        """
        return [Shard(number, shard) for number, shard in self._data['shards'].iteritems()]



__author__ = 'Joe Linn'

#import pylastica
import pylastica.response

class Response(pylastica.response.Response):
    def __init__(self, response_data, action, op_type):
        """

        @param response_data:
        @type response_data: dict or str
        @param action:
        @type action: pylastica.bulk.action.Action
        @param op_type: bulk operation type
        @type op_type: str
        """
        assert isinstance(action, pylastica.bulk.action.Action), "action must be an instance of Action: %r" % action
        super(Response, self).__init__(response_data)
        self._action = action
        self._op_type = op_type

    @property
    def action(self):
        """

        @return:
        @rtype: pylastica.bulk.action.Action
        """
        return self._action

    @property
    def op_type(self):
        """

        @return:
        @rtype: str
        """
        return self._op_type



__author__ = 'Joe Linn'

import base64
import pylastica.param
from pylastica.bulk.action import Action


class Document(pylastica.param.Param):
    OP_TYPE_CREATE = Action.OP_TYPE_CREATE

    def __init__(self, doc_id=None, data=None, doc_type=None, index=None):
        """

        @param doc_id:
        @type doc_id:
        @param data:
        @type data:
        @param doc_type:
        @type doc_type:
        @param index:
        @type index:
        @return:
        @rtype:
        """
        super(Document, self).__init__()
        self._data = {}
        self._script = None
        self._auto_populate = False
        self._doc_as_upsert = False
        self._upsert = None
        self.doc_id = doc_id
        if data is not None:
            self.data = data
        self.doc_type = doc_type
        self.index = index

    @property
    def doc_id(self):
        """
        Get the document id
        @return:
        @rtype: str
        """
        return self.get_param('_id') if self.has_param('_id') else ''

    @doc_id.setter
    def doc_id(self, doc_id):
        """
        Set the document id
        @param doc_id:
        @type doc_id: str
        """
        if doc_id is not None:
            self.set_param('_id', str(doc_id))

    def has_id(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_id') and len(self.doc_id) > 0

    def get(self, key):
        """
        Retrieve a piece of data from this document
        @param key:
        @type key: str
        @return:
        @rtype: mixed
        """
        if not self.has(key):
            raise pylastica.exception.InvalidException("Field %s does not exist." % key)
        return self._data[key]

    def set(self, key, value):
        """
        Set a field
        @param key:
        @type key:  str
        @param value:
        @type value: mixed
        @return:
        @rtype: self
        """
        assert isinstance(self._data, dict), "Document data is serialized data. Data creation is forbidden."
        self._data[key] = value
        return self

    def has(self, key):
        """
        Determine if a field exists in this document
        @param key:
        @type key: str
        @return:
        @rtype: bool
        """
        return isinstance(self._data, dict) and key in self._data

    def remove(self, key):
        """
        Remove a field from the document
        @param key:
        @type key: str
        @return:
        @rtype: self
        """
        if not self.has(key):
            raise pylastica.exception.InvalidException("Field %s does not exist." % key)
        del self._data[key]
        return self

    def add(self, key, value):
        """
        Add the given key / value pair to the document
        @param key:
        @type key: str
        @param value:
        @type value: mixed
        @return:
        @rtype: self
        """
        return self.set(key, value)

    def add_file(self, key, file_path, mime_type=None):
        """
        Adds a file to the index. The attachments plugin must be installed before using this feature.
        ./bin/plugin -install elasticsearch/elasticsearch-mapper-attachments/1.6.0
        @see http://tika.apache.org/0.7/formats.html
        @param key: key of file
        @type key: str
        @param file_path: path of the file
        @type file_path: str
        @param mime_type: optional header MIME type
        @type mime_type: str
        @return:
        @rtype: self
        """
        value = base64.b64encode(open(file_path, 'rb').read())
        if mime_type is not None:
            value = {
                '_content_type': mime_type,
                '_name': file_path,
                'content': value
            }
        return self.set(key, value)

    def add_file_content(self, key, content):
        """
        Add file content to the document
        @param key:
        @type key: str
        @param content: raw file content
        @type content: str
        @return:
        @rtype: self
        """
        return self.set(key, base64.b64encode(content))

    def add_geopoint(self, key, lat, lon):
        """
        Add a geopoint to the document
        @param key:
        @type key: str
        @param lat: latitude
        @type lat: float
        @param lon: longitude
        @type lon: float
        @return:
        @rtype: self
        """
        return self.set(key, {
            'lat': lat,
            'lon': lon
        })

    @property
    def data(self):
        """

        @return:
        @rtype: dict
        """
        return self._data

    @data.setter
    def data(self, data):
        """
        Overwrites the current document data with the given data
        @param data:
        @type data: dict or str
        """
        self._data = data

    def set_data(self, data):
        """
        Overwrites the current document data with the given data
        @param data:
        @type data: dict or str
        @return:
        @rtype: self
        """
        self._data = data
        return self

    @property
    def ttl(self):
        """
        Get the document's ttl
        @return:
        @rtype: str or int
        """
        return self.get_param('_ttl')

    @ttl.setter
    def ttl(self, ttl):
        """
        Set the ttl
        @param ttl:
        @type ttl: str or int
        """
        self.set_param('_ttl', ttl)

    def has_ttl(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_ttl')

    @property
    def doc_type(self):
        """
        Return document type
        @return:
        @rtype: str
        """
        return self.get_param('_type')

    @doc_type.setter
    def doc_type(self, doc_type):
        """
        Set document type
        @param doc_type:
        @type doc_type: str or pylastica.doc_type.DocType
        """
        if isinstance(doc_type, pylastica.doc_type.DocType):
            self.index = doc_type.index
            doc_type = doc_type.name
        self.set_param('_type', doc_type)

    @property
    def index(self):
        """
        Return the document's index name
        @return:
        @rtype: str
        """
        return self.get_param('_index')

    @index.setter
    def index(self, index):
        """
        Set the document's index name
        @param index:
        @type index: str or pylastica.index.Index
        """
        if isinstance(index, pylastica.index.Index):
            index = index.name
        self.set_param('_index', index)

    @property
    def version(self):
        """
        Get the document's version
        @return:
        @rtype: int
        """
        return self.get_param('_version')

    @version.setter
    def version(self, version):
        """
        Set the version of the document for use with optimistic concurrency control
        @param version:
        @type version:  int
        """
        self.set_param('_version', int(version))

    def has_version(self):
        """
        Determine whether or not the document has a version
        @return:
        @rtype: bool
        """
        return self.has_param('_version')

    @property
    def version_type(self):
        """
        Returns the document's version type
        @return:
        @rtype: str or int
        """
        return self.get_param('_version_type')

    @version_type.setter
    def version_type(self, version_type):
        """
        Set the document's version type
        @see: http://www.elasticsearch.org/guide/reference/api/index_.html
        @param version_type:
        @type version_type: int
        """
        self.set_param('_version_type', version_type)

    def has_version_type(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_version_type')

    @property
    def parent(self):
        """
        Returns the parent document's id
        @return:
        @rtype: str
        """
        return self.get_param('_parent')

    @parent.setter
    def parent(self, parent):
        """
        Set the parent document's id
        @param parent:
        @type parent: str or int
        """
        self.set_param('_parent', parent)

    def has_parent(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_parent')

    @property
    def op_type(self):
        """
        Return the operation type
        @return:
        @rtype: str
        """
        return self.get_param('_op_type')

    @op_type.setter
    def op_type(self, op_type):
        """
        Set the operation type
        @param op_type:
        @type op_type: str
        """
        self.set_param('_op_type', op_type)

    def has_op_type(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_op_type')

    @property
    def percolate(self):
        """
        Return percolate parameter
        @return:
        @rtype: str
        """
        return self.get_param('_percolate')

    @percolate.setter
    def percolate(self, percolate='*'):
        """
        Set percolate parameter
        @param percolate:
        @type percolate: str
        """
        self.set_param('_percolate', percolate)

    def has_percolate(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_percolate')

    @property
    def routing(self):
        """
        Return the routing for this document
        @return:
        @rtype: str
        """
        return self.get_param('_routing')

    @routing.setter
    def routing(self, routing):
        """
        Set routing for this document
        @param routing:
        @type routing: str
        """
        self.set_param('_routing', routing)

    def has_routing(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_routing')

    @property
    def fields(self):
        """

        @return:
        @rtype: str
        """
        return self.get_param('_fields')

    @fields.setter
    def fields(self, fields):
        """
        Set document fields
        @param fields:
        @type fields: str or list of str
        """
        if isinstance(fields, list):
            fields = ','.join(fields)
        self.set_param('_fields', fields)

    def has_fields(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_fields')

    def set_fields_source(self):
        """

        @return:
        @rtype: self
        """
        self.fields = '_source'
        return self

    @property
    def retry_on_conflict(self):
        """

        @return:
        @rtype: int
        """
        return self.get_param('_retry_on_conflict')

    @retry_on_conflict.setter
    def retry_on_conflict(self, retry):
        """

        @param retry:
        @type retry: int
        """
        self.set_param('_retry_on_conflict', int(retry))

    def has_retry_on_conflict(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_retry_on_conflict')

    @property
    def timestamp(self):
        """
        Return the document's timestamp
        @return:
        @rtype: int
        """
        return self.get_param('_timestamp')

    @timestamp.setter
    def timestamp(self, timestamp):
        """
        Set the document's timestamp
        @param timestamp:
        @type timestamp: int
        """
        self.set_param('_timestamp', timestamp)

    def has_timestamp(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_timestamp')

    @property
    def refresh(self):
        """

        @return:
        @rtype: bool
        """
        return self.get_param('_refresh')

    @refresh.setter
    def refresh(self, refresh=True):
        """

        @param refresh:
        @type refresh: bool
        """
        self.set_param('_refresh', bool(refresh))

    def has_refresh(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_refresh')

    @property
    def timeout(self):
        """

        @return:
        @rtype: str
        """
        return self.get_param('_timeout')

    @timeout.setter
    def timeout(self, timeout):
        """

        @param timeout:
        @type timeout: str
        """
        self.set_param('_timeout', timeout)

    def has_timeout(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_timeout')

    @property
    def consistency(self):
        """

        @return:
        @rtype: str
        """
        return self.get_param('_consistency')

    @consistency.setter
    def consistency(self, consistency):
        """

        @param consistency:
        @type consistency: str
        """
        self.set_param('_consistency', consistency)

    def has_consistency(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_consistency')

    @property
    def replication(self):
        """

        @return:
        @rtype: str
        """
        return self.get_param('_replication')

    @replication.setter
    def replication(self, replication):
        """

        @param replication:
        @type replication: str
        """
        self.set_param('_replication', replication)

    def has_replication(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_replication')

    @property
    def script(self):
        """

        @return:
        @rtype: pylastica.script.Script
        """
        raise pylastica.exception.NotImplementedException("The script getter is no longer available.")

    @script.setter
    def script(self, script):
        """

        @param script:
        @type script: pylastica.script.Script or dict or str
        """
        raise pylastica.exception.NotImplementedException("The script setter is no longer available.")

    @property
    def upsert(self):
        """

        @return:
        @rtype: dict
        """
        return self._upsert

    @upsert.setter
    def upsert(self, upsert):
        """

        @param upsert:
        @type upsert: dict
        """
        document = Document.create(upsert)
        self._upsert = document

    def has_upsert(self):
        """

        @return:
        @rtype: bool
        """
        return self._upsert is not None

    def has_script(self):
        """

        @return:
        @rtype: bool
        """
        raise pylastica.exception.NotImplementedException("has_script() is no longer available.")

    @property
    def doc_as_upsert(self):
        """

        @return:
        @rtype: bool
        """
        return self._doc_as_upsert

    @doc_as_upsert.setter
    def doc_as_upsert(self, value):
        """

        @param value:
        @type value: bool
        """
        self._doc_as_upsert = bool(value)

    @property
    def auto_populate(self):
        """

        @return:
        @rtype: bool
        """
        return self._auto_populate

    @auto_populate.setter
    def auto_populate(self, auto_populate):
        """

        @param auto_populate:
        @type auto_populate: bool
        """
        self._auto_populate = bool(auto_populate)

    def to_dict(self):
        """
        Return the document as a dict
        @return:
        @rtype: dict
        """
        doc = self.params
        doc['_source'] = self.data
        return doc

    def get_options(self, fields=None, with_underscore=False):
        """

        @param fields: if None, all options will be returned
        @type fields: list of str
        @param with_underscore: determines whether or not option keys should contaion underscore prefix
        @type with_underscore: bool
        @return:
        @rtype: dict
        """
        if fields is not None:
            data = {}
            for field in fields:
                key = '_' + field.lstrip('_')
                if self.has_param(key) and str(self.get_param(key)) != '':
                    data[key] = self.get_param(key)
        else:
            data = self.params
        if not with_underscore:
            data = {key.lstrip('_'): value for key, value in data.iteritems()}
        return data

    @classmethod
    def create(cls, data):
        """

        @param cls:
        @type cls:
        @param data:
        @type data: dict or Document
        @return:
        @rtype: Document
        """
        if isinstance(data, cls):
            return data
        elif isinstance(data, dict):
            return cls('', data)
        else:
            raise pylastica.exception.InvalidException("Failed to create document. Invalid data provided: %r" % data)


__author__ = 'Joe Linn'

import pylastica.aggregation.abstractaggregation as abstract


class GeohashGrid(abstract.AbstractAggregation):
    def __init__(self, name, field):
        """
        @param name: the name of this aggregation
        @type name: str
        @param field: the field on which to perform this aggregation
        @type field: str
        """
        super(GeohashGrid, self).__init__(name)
        self.set_field(field)

    def set_field(self, field):
        """
        Set the field for this aggregation
        @param field: the name of the document field on which to perform this aggregation
        @type field: str
        @return:
        @rtype: GeohashGrid
        """
        return self.set_param("field", field)

    def set_precision(self, precision):
        """
        Set the precision for this aggregation
        @param precision: an integer between 1 and 12, inclusive. Defaults to 5.
        @type precision: int
        @return:
        @rtype: GeohashGrid
        """
        return self.set_param("precision", precision)

    def set_size(self, size):
        """
        Set the maximum number of buckets to return
        @param size: defaults to 10,000
        @type size: int
        @return:
        @rtype: GeohashGrid
        """
        return self.set_param("size", size)

    def set_shard_size(self, shard_size):
        """
        Set the number of results returned from each shard
        @param shard_size:
        @type shard_size: int
        @return:
        @rtype: GeohashGrid
        """
        return self.set_param("shard_size", shard_size)


__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter
import pylastica.exception

class GeoPolygon(AbstractFilter):
    def __init__(self, key, points):
        """
        @param key:
        @type key: str
        @param points: points making up the polygon [{'lat': float, 'lon': float}]
        @type points: list of list or list of dict
        """
        super(GeoPolygon, self).__init__()
        self._key = key
        self._points = points

    def to_dict(self):
        """
        @rtype : dict
        """
        return {
            'geo_polygon':{
                self._key:{
                    'points': self._points
                }
            }
        }





__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter

class Query(AbstractFilter):
    def __init__(self, query=None):
        """

        @param query: query
        @type query: dict or pylastica.query.abstract.AbstractQuery
        """
        super(Query, self).__init__()
        if query is not None:
            self.set_query(query)

    def set_query(self, query):
        """
        Set the query
        @param query:
        @type query: dict or pylastica.query.abstract.AbstractQuery
        @return:
        @rtype: self
        """
        if not isinstance(query, pylastica.query.abstract.AbstractQuery) and not isinstance(query, dict):
            raise TypeError("query must be of type dict or AbstractQuery: %r" % query)
        if isinstance(query, pylastica.query.abstract.AbstractQuery):
            query = query.to_dict()
        self._query = query
        return self

    def _get_base_name(self):
        """

        @return:
        @rtype: str
        """
        if self._params is None or not len(self._params):
            return 'query'
        else:
            return 'fquery'

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        data = super(Query, self).to_dict()
        name = self._get_base_name()
        filter_data = data[name]
        if not filter_data:
            filter_data = self._query
        else:
            filter_data['query'] = self._query
        data[name] = filter_data
        return data




__author__ = 'Joe Linn'

from .abstractfacet import AbstractFacet
import pylastica.exception


class GeoDistance(AbstractFacet):
    """
    @see: http://www.elasticsearch.org/guide/reference/api/search/facets/geo-distance-facet/
    """

    def set_ranges(self, ranges):
        """
        Set all ranges for the facet
        @param ranges: [{'to': 50}, {'to': 40, 'from': 20}]
        @type ranges: list of dict
        @return:
        @rtype: self
        """
        return self.set_param('ranges', ranges)

    def add_range(self, range_from=None, range_to=None):
        """
        Add a range to the facet
        @param range_from: optional. Assumed to be boundless if omitted.
        @type range_from: int
        @param range_to: optional. Assumed to be boundless if omitted.
        @type range_to: int
        @return:
        @rtype: self
        """
        if range_from is None and range_to is None:
            raise pylastica.exception.InvalidException("At least one of range_from and range_to must be set.")
        range_dict = {}
        if range_from is not None:
            range_dict['from'] = range_from
        if range_to is not None:
            range_dict['to'] = range_to
        return self.add_param('ranges', range_dict)

    def set_geo_point(self, type_field, lat, lon):
        """
        Set the relative GeoPoint for this facet
        @param type_field: index type and field (pin.location or foo.bar, for example)
        @type type_field: str
        @param lat: latitude
        @type lat: float
        @param lon: longitude
        @type lon: float
        @return:
        @rtype: self
        """
        return self.set_param(type_field, {'lat': lat, 'lon': lon})

    def set_value_field(self, field):
        """
        Set the value field for this facet
        @param field: field name
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param('value_field', field)

    def set_value_script(self, script):
        """
        Set a script for this facet
        @param script:
        @type script: pylastica.script.Script
        @return:
        @rtype: self
        """
        if not isinstance(script, pylastica.script.Script):
            raise TypeError("script must be an instance of Script: %r" % script)
        return self.set_param('value_script', script.script).set_param('params', script.params)

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        self._set_facet_param('geo_distance', self._params)
        return super(GeoDistance, self).to_dict()





__author__ = 'Joe Linn'

from .abstract import AbstractQuery
from pylastica.filter.abstractfilter import AbstractFilter


class FunctionScore(AbstractQuery):
    """
    @see http://www.elasticsearch.org/guide/reference/query-dsl/function-score-query/
    """

    BOOST_MODE_MULTIPLY = 'multiply'
    BOOST_MODE_REPLACE = 'replace'
    BOOST_MODE_SUM = 'sum'
    BOOST_MODE_AVERAGE = 'avg'
    BOOST_MODE_MAX = 'max'
    BOOST_MODE_MIN = 'min'

    SCORE_MODE_MULTIPLY = 'multiply'
    SCORE_MODE_SUM = 'sum'
    SCORE_MODE_AVERAGE = 'avg'
    SCORE_MODE_FIRST = 'first'
    SCORE_MODE_MAX = 'max'
    SCORE_MODE_MIN = 'min'

    DECAY_GAUSS = 'gauss'
    DECAY_EXPONENTIAL = 'exp'
    DECAY_LINEAR = 'linear'

    def __init__(self):
        super(FunctionScore, self).__init__()
        self._functions = []

    def set_query(self, query):
        """
        Set the query for this function_score query
        @param query: a Query object
        @type query: pylastica.query.AbstractQuery
        @return:
        @rtype: self
        """
        if not isinstance(query, AbstractQuery):
            raise TypeError("query must be and instance of a descendant of AbstractQuery: %r" % query)
        self.set_param('query', query.to_dict())

    def set_filter(self, filter):
        """
        Set the filter for this function_score query
        @param filter:
        @type filter: AbstractFilter
        @return:
        @rtype: self
        """
        return self.set_param('filter', filter.to_dict())

    def add_function(self, function_type, function_params, filter=None, boost=None):
        """
        Add a function to the function_score query
        @param function_type: valid values are gauss, exp, linear, script_score
        @type function_type: str
        @param function_params: the body of the function. See documentation for proper syntax.
        @type function_params: dict
        @param filter: an optional filter to apply to the function
        @type filter: pylastica.filter.AbstractFilter
        @param boost: optional boost value associated with this function
        @type boost: float
        @return:
        @rtype: self
        """
        function = {
            function_type: function_params
        }
        if filter is not None:
            if not isinstance(filter, AbstractFilter):
                raise TypeError("filter must be an instance of a descendant of AbstractFilter: %r" % filter)
            function['filter'] = filter.to_dict()
        if boost is not None:
            function['boost'] = float(boost)
        self._functions.append(function)
        return self

    def add_script_score_function(self, script, filter=None, boost=None):
        """
        Add a script_score function to the query
        @param script: a Script object
        @type script: pylastica.Script
        @param filter: a filter object
        @type filter: AbstractFilter
        @param boost: optional boost value associated with this function
        @type boost: float
        @return:
        @rtype: self
        """
        return self.add_function('script_score', script.to_dict(), filter, boost)

    def add_decay_function(self, function, field, origin, scale, offset=None, decay=None, scale_weight=None,
                           filter=None, boost=None):
        """
        Add a decay function to the query
        @param function: the decay function type. See DECAY_* properties for valid options.
        @type function: str
        @param field: the document field on which to perform the decay function
        @type field: str
        @param origin: the origin value for this decay function
        @type origin: str
        @param scale: a scale to define the rate of decay for this function
        @type scale: str
        @param offset: optional. If defined, this function will only be computed for documents with a distance greater than this value.
        @type offset: str
        @param decay: optionally defines how documents are scored at the distance given in the scale parameter
        @type decay: float
        @param scale_weight: an optional factor by which to multiply the score at the value provided by the scale parameter
        @type scale_weight: float
        @param filter: a filter associated with this function
        @type filter: AbstractFilter
        @param boost: an optional boost value associated with this function
        @type boost: float
        @return:
        @rtype: self
        """
        valid_decay = [
            self.DECAY_GAUSS,
            self.DECAY_EXPONENTIAL,
            self.DECAY_LINEAR
        ]
        if function not in valid_decay:
            raise ValueError("%s is not a supported decay function." % function)
        function_params = {
            field: {
                'origin': origin,
                'scale': scale
            }
        }
        if offset is not None:
            function_params[field]['offset'] = offset
        if decay is not None:
            function_params[field]['decay'] = float(decay)
        if scale_weight is not None:
            function_params[field]['scale_weight'] = float(scale_weight)
        return self.add_function(function, function_params, filter, boost)

    def set_boost(self, boost=1.0):
        """
        Set the boost for the query
        @param boost:
        @type boost: float
        @return:
        @rtype: self
        """
        return self.set_param('boost', float(boost))

    def set_max_boost(self, max_boost):
        """
        Restrict the combined boost of the function_score query and its child query
        @param max_boost:
        @type max_boost: float
        @return:
        @rtype: self
        """
        return self.set_param('max_boost', float(max_boost))

    def set_boost_mode(self, mode):
        """
        The boost mode determines how the score of this query is combined with that of the child query
        @param mode: see BOOST_MODE_* properties for valid options. Default is multiply.
        @type mode: str
        @return:
        @rtype: self
        """
        valid_boost_modes = [
            self.BOOST_MODE_MULTIPLY,
            self.BOOST_MODE_REPLACE,
            self.BOOST_MODE_SUM,
            self.BOOST_MODE_AVERAGE,
            self.BOOST_MODE_MAX,
            self.BOOST_MODE_MIN
        ]
        if mode not in valid_boost_modes:
            raise ValueError("%s is not a valid boost mode." % mode)
        return self.set_param('boost_mode', mode)

    def set_random_score(self, seed=None):
        """
        If set, this query will return results in random order.
        @param seed: Set a seed value to return results in the same random order for consistent pagination.
        @type seed: int
        @return:
        @rtype: self
        """
        seed_param = {}
        if seed is not None:
            seed_param['seed'] = int(seed)
        return self.set_param('random_score', seed_param)

    def set_score_mode(self, mode):
        """
        Set score method
        @param mode: see SCORE_MODE_* properties for options. Default is multiply.
        @type mode: str
        @return:
        @rtype: self
        """
        valid_score_modes = [
            self.SCORE_MODE_MULTIPLY,
            self.SCORE_MODE_SUM,
            self.SCORE_MODE_AVERAGE,
            self.SCORE_MODE_FIRST,
            self.SCORE_MODE_MAX,
            self.SCORE_MODE_MIN
        ]
        if mode not in valid_score_modes:
            raise ValueError("%s is not a valid score mode." % mode)
        return self.set_param('score_mode', mode)

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        if len(self._functions):
            self.set_param('functions', self._functions)
        return super(FunctionScore, self).to_dict()



__author__ = 'Joe Linn'

import pylastica

class Mapping(object):
    def __init__(self, doc_type=None, properties=None):
        """

        @param type: optional DocType object
        @type type: pylastica.doc_type.DocType
        @param properties:
        @type properties: dict
        """
        self._mapping = {}
        self._type = None
        if doc_type:
            self.doc_type = doc_type
        if properties:
            self.set_properties(properties)

    @property
    def doc_type(self):
        """
        Return the document type object
        @return:
        @rtype: pylastica.doc_type.DocType
        """
        return self._type

    @doc_type.setter
    def doc_type(self, doc_type):
        """

        @param doc_type:
        @type doc_type: pylastica.doc_type.DocType
        """
        assert isinstance(doc_type, pylastica.doc_type.DocType), "doc_type jmust be of type DocType: %r" % doc_type
        self._type = doc_type

    def set_properties(self, properties):
        """
        Set mapping properties
        @param properties:
        @type properties: dict
        @return:
        @rtype: self
        """
        return self.set_param('properties', properties)

    def set_meta(self, meta):
        """
        Set mapping meta
        @param meta:
        @type meta: dict
        @return:
        @rtype: self
        @see: http://www.elasticsearch.org/guide/reference/mapping/meta/
        """
        return self.set_param('_meta', meta)

    def set_source(self, source):
        """
        Sets source values
        @param source:
        @type source: dict
        @return:
        @rtype: self
        """
        return self.set_param('_source', source)

    def disable_source(self, enabled=False):
        """
        Disables the source in the index.
        @param enabled:
        @type enabled: bool
        @return:
        @rtype: self
        """
        return self.set_source({'enabled': enabled})

    def set_param(self, key, value):
        """
        Sets raw parameters.  See ES docs for options.
        @param key: key name
        @type key: str
        @param value:
        @type value: mixed
        @return:
        @rtype: self
        """
        self._mapping[key] = value
        return self

    def set_all_field(self, params):
        """
        Set params for the '_all' field
        @param params:
        @type params: dict
        @return:
        @rtype: self
        """
        return self.set_param('_all', params)

    def enable_all_field(self, enabled=True):
        """
        Enable (or disable) the '_all' field
        @param enabled: Enabled if True, disabled if False. Defaults to True.
        @type enabled: bool
        @return:
        @rtype: self
        """
        return self.set_all_field({'enabled': enabled})

    def set_ttl(self, params):
        """
        Set TTL
        @param params: ttl parameters
        @type params: dict
        @return:
        @rtype: self
        """
        return self.set_param('_ttl', params)

    def enable_ttl(self, enabled=True):
        """
        Enable TTL for all documents of this type
        @param enabled:
        @type enabled: bool
        @return:
        @rtype: self
        """
        return self.set_ttl({'enabled': enabled})

    def set_parent(self, doc_type):
        """
        Set parent doc type
        @param doc_type:
        @type doc_type: str
        @return:
        @rtype: self
        """
        return self.set_param('_parent', {'type': doc_type})

    def to_dict(self):
        """
        Convert the mapping to a dict
        @return:
        @rtype: dict
        """
        doc_type = self.doc_type
        if doc_type is None:
            raise pylastica.exception.InvalidException("doc_type must be set.")
        return {doc_type.name: self._mapping}

    def send(self):
        """
        Sends the mapping to the server
        @return:
        @rtype: pylastica.response.Response
        """
        return self.doc_type.request('_mapping', pylastica.request.Request.PUT, self.to_dict())

    @classmethod
    def create(cls, mapping):
        """
        Create a mapping object
        @param cls:
        @type cls:
        @param mapping: mapping object or properties dict
        @type mapping: pylastica.doc_type.Mapping or dict
        @return:
        @rtype: pylastica.doc_type.Mapping
        """
        if isinstance(mapping, dict):
            mapping_object = cls()
            mapping_object.set_properties(mapping)
        else:
            mapping_object = mapping
        assert isinstance(mapping_object, cls), "Invalid object type"
        return mapping_object



from __future__ import absolute_import
__author__ = 'Joe Linn'
__docformat__ = 'epytext'

from . import aggregation
from . import bulk
from .client import *
from . import cluster
from .connection import *
from .document import *
from . import doc_type
from . import exception
from . import facet
from . import filter
from . import index
# from .log import *
from . import node
# from .param import *
from .percolator import *
from . import query
from .request import *
from .response import *
from .resultset import *
from . import river
from .script import *
from .scriptfields import *
from .search import *
from .status import *
from . import suggest
# from .searchable import *
# from . import transport
# from . import util



__author__ = 'Joe Linn'

import pylastica.aggregation.abstractaggregation as abstract


class GeoDistance(abstract.AbstractAggregation):
    def __init__(self, name, field, origin):
        """
        @param name: the name of this aggregation
        @type name: str
        @param field: the field on which to perform this aggregation
        @type field: str
        @param origin: the point from which distances will be calculated
        @type origin: dict or str or list
        """
        super(GeoDistance, self).__init__(name)
        self.set_field(field).set_origin(origin)

    def set_field(self, field):
        """
        Set the field for this aggregation
        @param field: the name of the document field on which to perform this aggregation
        @type field: str
        @return:
        @rtype: GeoDistance
        """
        return self.set_param("field", field)

    def set_origin(self, origin):
        """
        Set the origin point from which distances will be calculated
        @param origin: valid formats are { "lat" : 52.3760, "lon" : 4.894 }, "52.3760, 4.894", and [4.894, 52.3760]
        @type origin: dict or str or list
        @return:
        @rtype: GeoDistance
        """
        return self.set_param("origin", origin)

    def add_range(self, from_value=None, to_value=None):
        """
        Add a distance range to this aggregation
        @param from_value: A distance
        @type from_value: int
        @param to_value: A distnace
        @type to_value: int
        @return:
        @rtype: GeoDistance
        """
        if from_value is None and to_value is None:
            raise ValueError("Either from_value or to_value must be set. Both cannot be None.")
        range = {}
        if from_value is not None:
            range['from'] = from_value
        if to_value is not None:
            range['to'] = to_value
        return self.add_param('ranges', range)

    def set_unit(self, unit):
        """
        Set the unit of distance measure for this aggregation
        @param unit: defaults to km
        @type unit: str
        @return:
        @rtype: GeoDistance
        """
        return self.set_param("unit", unit)

    def set_distance_type(self, distance_type):
        """
        Set the method by which distances will be calculated
        @param distance_type: valid options are sloppy_arc (default), arc, and plane
        @type distance_type: str
        @return:
        @rtype: GeoDistance
        """
        return self.set_param("distance_type", distance_type)


__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
import pylastica.doc_type
from .abstractfilter import AbstractFilter

class Ids(AbstractFilter):
    def __init__(self, doc_type=None, ids=None):
        """

        @param doc_type: document type
        @type doc_type: str or pylastica.doc_type.DocType
        @param ids: list of document ids
        @type ids: list of str or str
        """
        super(Ids, self).__init__()
        self.set_doc_type(doc_type)
        self.set_ids(ids)

    def add_id(self, doc_id):
        """
        Add a document id to the filter
        @param id: document id
        @type id: str
        @return:
        @rtype: self
        """
        return self.add_param('values', doc_id)

    def add_doc_type(self, doc_type):
        """
        Add another document type to the filter
        @param doc_type: document type
        @type doc_type: str or pylastica.doc_type.DocType
        @return:
        @rtype: self
        """
        if isinstance(doc_type, pylastica.doc_type.DocType):
            doc_type = doc_type.name
        return self.add_param('type', doc_type)

    def set_doc_type(self, doc_type):
        """
        Set the document type
        @param doc_type: doc type
        @type doc_type: str or pylastica.doc_type.DocType or list of str
        @return:
        @rtype: self
        """
        if isinstance(doc_type, pylastica.doc_type.DocType):
            doc_type = doc_type.name
        return self.set_param('type', doc_type)

    def set_ids(self, ids):
        """
        Set document ids for the filter
        @param ids:
        @type ids: list of str or str
        @return:
        @rtype: self
        """
        if not isinstance(ids, list):
            ids = [ids]
        return self.set_param('values', ids)



__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter
import pylastica.exception

class GeoBoundingBox(AbstractFilter):
    def __init__(self, key, coordinates):
        """
        @param key:
        @type key: str
        @param coordinates: list with top left coordinates as first index, and bottom right coordinates as second index
        @type coordinates: list
        """
        super(GeoBoundingBox, self).__init__()
        self.add_coordinates(key, coordinates)

    def add_coordinates(self, key, coordinates):
        """
        Add coordinates
        @param key:
        @type key: str
        @param coordinates: list with top left coordinates as first index, and bottom right coordinates as second index
        @type coordinates: list
        @return:
        @rtype: self
        """
        if len(coordinates) != 2:
            raise pylastica.exception.InvalidException("Expected coordinates to be a list with two elements.")
        self.set_param(key, {
            'top_left': coordinates[0],
            'bottom_right': coordinates[1]
        })
        return self



__author__ = 'Joe Linn'

from .abstractfacet import AbstractFacet

class Histogram(AbstractFacet):
    def set_field(self, field):
        """
        Set the field for the histogram
        @param field: field name
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param('field', field)

    def set_interval(self, interval):
        """
        Set the interval value
        @param interval:
        @type interval: str
        @return:
        @rtype: self
        """
        return self.set_param('interval', interval)

    def set_key_value_fields(self, key_field, value_field):
        """
        Set key_field and value_field
        @param key_field:
        @type key_field: str
        @param value_field:
        @type value_field: str
        @return:
        @rtype: self
        """
        return self.set_param('key_field', key_field).set_param('value_field', value_field)

    def set_key_value_scripts(self, key_script, value_script):
        """
        Set key and value scripts
        @param key_script:
        @type key_script: str
        @param value_script:
        @type value_script: str
        @return:
        @rtype: self
        """
        return self.set_param('key_script', key_script).set_param('value_script', value_script)

    def set_script_params(self, params):
        """
        Set params to be used with scripts
        @param params:
        @type params: dict
        @return:
        @rtype: self
        """
        return self.set_param('params', params)

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        self._set_facet_param('histogram', self._params)
        return super(Histogram, self).to_dict()





__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter

class Missing(AbstractFilter):
    def __init__(self, field=''):
        """

        @param field: field name
        @type field: str
        """
        super(Missing, self).__init__()
        if len(field):
            self.set_field(field)

    def set_field(self, field):
        """
        Set the field
        @param field: document field
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param('field', str(field))



__author__ = 'Joe Linn'

from .abstractfilter import AbstractMulti


class BoolAnd(AbstractMulti):
    """
    @see: @link http://www.elasticsearch.org/guide/reference/query-dsl/and-filter.html
    """
    def _get_base_name(self):
        return 'and'



__author__ = 'Joe Linn'


class Stats(object):
    def __init__(self, node):
        """
        @param node:
        @type node: pylastica.node.Node
        """
        super(Stats, self).__init__()
        import pylastica.node
        assert isinstance(node, pylastica.node.Node), "node must be an instance of pylastica.node.Node: %r" % node
        self._node = node
        self.refresh()

    def get(self, *args):
        """
        Returns an entry in the data array based on the args.
        Example: get('index', 'test', 'example')
        @param args:
        @type args: str
        @return: returns None if the requested data cannot be found
        @rtype: str or dict or None
        """
        data = self.data
        for arg in args:
            if arg in data:
                data = data[arg]
            else:
                return None
        return data

    @property
    def data(self):
        """
        Returns all stats data
        @return:
        @rtype: dict
        """
        return self._data

    @property
    def node(self):
        """
        Returns the node object
        @return:
        @rtype: pylastica.node.Node
        """
        return self._node

    @property
    def response(self):
        """
        Returns the Response object
        @return:
        @rtype: pylastica.response.Response
        """
        return self._response

    def refresh(self):
        """
        Reloads all node information
        @return:
        @rtype: pylastica.response.Response
        """
        self._response = self.node.client.request('_nodes/%s/stats' % self.node.name)
        self._data = self.response.data['nodes']



__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter

class Limit(AbstractFilter):
    def __init__(self, limit):
        """
        @param limit: limit
        @type limit: int
        """
        super(Limit, self).__init__()
        self.set_limit(limit)

    def set_limit(self, limit):
        """
        Set the limit
        @param limit:
        @type limit: int
        @return:
        @rtype: self
        """
        return self.set_param('value', int(limit))



__author__ = 'Joe Linn'

from .abstract import AbstractQuery
import pylastica.exception


class Terms(AbstractQuery):
    def __init__(self, key='', terms=None):
        """
        Construct terms query
        @param key: terms key
        @type key: str
        @param terms: terms list
        @type terms: list
        """
        super(Terms, self).__init__()
        self.set_terms(key, terms)

    def set_terms(self, key, terms):
        """
        Set key and terms for the query
        @param key:
        @type key: str
        @param terms:
        @type terms: list
        @return:
        @rtype: self
        """
        self._key = key
        if terms is None:
            terms = []
        self._terms = terms
        return self

    def add_term(self, term):
        """
        Add a single term to the list
        @param term:
        @type term: str
        @return:
        @rtype: self
        """
        self._terms.append(term)
        return self

    def set_minimum_match(self, minimum):
        """
        Set the minimum matching value
        @param minimum:
        @type minimum: int
        @return:
        @rtype: self
        """
        return self.set_param('minimum_match', minimum)

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        if self._key is None:
            raise pylastica.exception.InvalidException("Terms key must be set.")
        self.set_param(self._key, self._terms)
        return super(Terms, self).to_dict()



__author__ = 'Joe Linn'

from .abstract import AbstractCandidateGenerator


class DirectGenerator(AbstractCandidateGenerator):
    """
    @see: http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/search-suggesters-phrase.html#_direct_generators
    """
    SUGGEST_MODE_MISSING = 'missing'
    SUGGEST_MODE_POPULAR = 'popular'
    SUGGEST_MODE_ALWAYS = 'always'

    def __init__(self, field):
        """

        @param field:
        @type field: str
        """
        super(DirectGenerator, self).__init__()
        self.set_field(field)

    def set_field(self, field):
        """
        Set the field name from which to fetch candidate suggestions
        @param field:
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param("field", field)

    def set_size(self, size):
        """
        Set the maximum corrections to be returned per suggest text token
        @param size:
        @type size: int
        @return:
        @rtype: self
        """
        return self.set_param("size", size)

    def set_suggest_mode(self, mode):
        """

        @param mode: see SUGGEST_MODE_* constants for options
        @type mode: str
        @return:
        @rtype: self
        """
        return self.set_param("suggest_mode", mode)

    def set_max_edits(self, max_edits):
        """

        @param max_edits: can only be a value between 1 and 2. Defaults to 2.
        @type max_edits: float
        @return:
        @rtype: self
        """
        return self.set_param("max_edits", max_edits)

    def set_prefix_length(self, length):
        """

        @param length: defaults to 1
        @type length: int
        @return:
        @rtype: self
        """
        return self.set_param("prefix_len", length)

    def set_min_word_length(self, min_length):
        """

        @param min_length: defaults to 4
        @type min_length: int
        @return:
        @rtype: self
        """
        return self.set_param("min_word_len", min_length)

    def set_max_inspections(self, max_inspections):
        """

        @param max_inspections:
        @type max_inspections: int
        @return:
        @rtype: self
        """
        return self.set_param("max_inspections", max_inspections)

    def set_min_doc_frequency(self, min_frequency):
        """

        @param min_frequency:
        @type min_frequency: float
        @return:
        @rtype: self
        """
        return self.set_param("min_doc_freq", min_frequency)

    def set_max_term_frequency(self, max_frequency):
        """

        @param max_frequency:
        @type max_frequency: float
        @return:
        @rtype: self
        """
        return self.set_param("max_term_freq", max_frequency)

    def set_pre_filter(self, pre):
        """
        Set an analyzer to be applied to the original token prior to candidate generation
        @param pre: an analyzer
        @type pre: str
        @return:
        @rtype: self
        """
        return self.set_param("pre_filter", pre)

    def set_post_filter(self, post):
        """
        Set an analyzer to be applied to generated tokens before they are passed to the phrase scorer
        @param post:
        @type post: str
        @return:
        @rtype: self
        """
        return self.set_param("post_filter", post)


__author__ = 'Joe Linn'

import pylastica.client
from .info import *
from .stats import *


class Node(object):
    def __init__(self, name, client):
        """
        @param name: node name
        @type name: str
        @param client:
        @type client: pylastica.client.Client
        """
        super(Node, self).__init__()
        assert isinstance(client, pylastica.client.Client), "client must be an instance of pylastica.client.Client: %r" % client
        self._name = name
        self._client = client
        self.refresh()

    @property
    def name(self):
        """
        Get the name ofthe node
        @return:
        @rtype: str
        """
        return self._name

    @property
    def client(self):
        """
        Get the client object
        @return:
        @rtype: pylastica.client.Client
        """
        return self._client

    @property
    def stats(self):
        """
        Return stats object for the current node
        @return:
        @rtype: pylastica.node.stats.Stats
        """
        if not self._stats:
            self._stats = Stats(self)
        return self._stats

    @property
    def info(self):
        """
        Return an info object for the current node
        @return:
        @rtype: pylastica.node.info.Info
        """
        if not self._info:
            self._info = Info(self)
        return self._info

    def refresh(self):
        """
        Refresh node information. This should be called after updating a node.
        @return:
        @rtype: self
        """
        self._stats = None
        self._info = None
        return self

    def shutdown(self, delay='1s'):
        """
        Shut down this node.
        @param delay: (optional) Delay after which the node will be shut down. Defaults to 1s
        @type delay: str
        @return:
        @rtype: pylastica.response.Response
        """
        path = "_cluster/nodes/%s/_shutdown" % self.name
        return self.client.request(path, pylastica.request.Request.POST, query={'delay': delay})



__author__ = 'Joe Linn'

import json
import thrift
import thrift.transport
import thrift.transport.TSocket
import thrift.protocol
import Rest
from pylastica.transport import AbstractTransport
import pylastica.response
import pylastica.exception.connection


class ThriftTransport(AbstractTransport):
    def __init__(self, connection=None):
        """
        @param connection:
        @type connection:  pylastica.connection.Connection
        """
        super(ThriftTransport, self).__init__(connection)
        self._clients = {}

    def _create_client(self, host, port, send_timeout=None, recv_timeout=None, framed_transport=False):
        """

        @param host: hostname / ip address
        @type host: str
        @param port:
        @type port: int
        @param send_timeout: milliseconds
        @type send_timeout: int
        @param recv_timeout: milliseconds
        @type recv_timeout: int
        @param framed_transport:
        @type framed_transport: bool
        @return:
        @rtype: Rest.Client
        """
        socket = thrift.transport.TSocket.TSocket(host, port)
        if send_timeout is not None:
            socket.setSendTimeout(send_timeout)
        if recv_timeout is not None:
            socket.setRecvTimeout(recv_timeout)
        if framed_transport:
            transport = thrift.transport.TTransport.TFramedTransport(socket)
        else:
            transport = thrift.transport.TTransport.TBufferedTransport(socket)
        protocol = thrift.protocol.TBinaryProtocol.TBinaryProtocolAccelerated(transport)
        client = Rest.Client(protocol)
        transport.open()
        return client

    def _get_client(self, host, port, send_timeout=None, recv_timeout=None, framed_transport=False):
        """

        @param host: hostname / ip address
        @type host: str
        @param port:
        @type port: int
        @param send_timeout: milliseconds
        @type send_timeout: int
        @param recv_timeout: milliseconds
        @type recv_timeout: int
        @param framed_transport:
        @type framed_transport: bool
        @return:
        @rtype: Rest.Client
        """
        key = "%s:%s" % (host, port)
        if key not in self._clients:
            self._clients[key] = self._create_client(host, port, send_timeout, recv_timeout, framed_transport)
        return self._clients[key]

    def execute(self, request, params):
        """
        Executes a transport request
        @param request: request object
        @type request: pylastica.request.Request
        @param params: hostname, port, path, etc.
        @type params: dict
        @return:
        @rtype: pylastica.response.Response
        """
        connection = self.connection
        send_timeout = int(connection.get_config('send_timeout')) if connection.has_config('send_timeout') else None
        recv_timeout = int(connection.get_config('recv_timeout')) if connection.has_config('recv_timeout') else None
        framed_transport = bool(connection.get_config('framed_transport')) if connection.has_config('framed_transport') else False
        try:
            client = self._get_client(connection.host, connection.port, send_timeout, recv_timeout, framed_transport)
            rest_request = Rest.RestRequest()
            rest_request.method = Rest.Method._NAMES_TO_VALUES[request.get_method()]
            rest_request.uri = request.path

            query = request.query
            if query is not None:
                rest_request.parameters = query
            data = request.get_data()
            if data is not None:
                if isinstance(data, dict):
                    content = json.dumps(data)
                else:
                    content = data
                rest_request.body = content
            result = client.execute(rest_request)
            response = pylastica.response.Response(result.body)
        except thrift.Thrift.TException as e:
            response = pylastica.response.Response('')
            raise pylastica.exception.connection.ThriftException(e, request, response)
        if response.has_error():
            raise pylastica.exception.ResponseException(request, response)
        return response



__author__ = 'Joe Linn'

import pylastica.searchable
from .settings import *
from .stats import *
from .status import *


class Index(pylastica.searchable.Searchable):
    def __init__(self, client, name):
        """

        @param client:
        @type client: pylastica.client.Client
        @param name: index name
        @type name: str
        """
        assert isinstance(client, pylastica.client.Client), "client must be an instance of Client: %r" % client
        name = str(name)
        assert isinstance(name, str), "name must be a str: %r" % name
        self._name = name
        self._client = client

    def get_doc_type(self, doc_type):
        """
        Return a DocType object for the current index with the given type name
        @param doc_type: type name
        @type doc_type: str
        @return:
        @rtype: pylastica.doc_type.DocType
        """
        return pylastica.doc_type.DocType(self, doc_type)

    @property
    def status(self):
        """

        @return:
        @rtype: pylastica.index.Status
        """
        return self.get_status()

    def get_status(self):
        """
        Returns the current status of the idnex
        @return:
        @rtype: pylastica.index.Status
        """
        return pylastica.index.Status(self)

    @property
    def stats(self):
        """

        @return:
        @rtype: pylastica.index.Stats
        """
        return self.get_stats()

    def get_stats(self):
        """
        Return index stats
        @return:
        @rtype: pylastica.index.Stats
        """
        return pylastica.index.Stats(self)

    @property
    def mapping(self):
        """
        Get all the type mappings for an index
        @return:
        @rtype: dict
        """
        return self.request('_mapping', pylastica.request.Request.GET).data

    @property
    def settings(self):
        """
        Return the index settings object
        @return:
        @rtype: pylastica.index.Settings
        """
        return pylastica.index.Settings(self)

    def add_documents(self, docs):
        """
        Use _bulk to send docs to the server
        @param docs:
        @type docs: list of pylastica.document.Document
        @return:
        @rtype: pylastica.bulk.ResponseSet
        """
        for doc in docs:
            doc.index = self.name
        return self.client.add_documents(docs)

    def delete(self):
        """
        Delete this index
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request('', pylastica.request.Request.DELETE)

    def optimize(self, args=None):
        """
        Optimizes a search index
        @see: http://www.elasticsearch.org/guide/reference/api/admin-indices-optimize.html
        @param args: optional
        @type args: dict
        @return:
        @rtype: dict
        """
        return self.request('_optimize', pylastica.request.Request.POST, args)

    def refresh(self):
        """
        Refresh the index
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request('_refresh', pylastica.request.Request.POST)

    def create(self, args=None, options=None):
        """
        Creates a new index with the given args
        @see: http://www.elasticsearch.org/guide/reference/api/admin-indices-create-index.html
        @param args: optional
        @type args: dict
        @param options: bool: deletes index first if it exists (defaults to False); dict: dictionary of options
        @type options: bool or dict
        @return: server response
        @rtype: pylastica.response.Response
        """
        path = ''
        query = {}
        if isinstance(options, bool):
            if options:
                try:
                    self.delete()
                except pylastica.exception.ResponseException:
                    #index doesn't exist yet
                    pass
        else:
            if isinstance(options, dict):
                for key, value in options.iteritems():
                    if key == 'recreate':
                        try:
                            self.delete()
                        except pylastica.exception.ResponseException:
                            #index doesn't exist yet
                            pass
                    elif key == 'routing':
                        query = {'routing': value}
                    else:
                        raise pylastica.exception.InvalidException("Invalid option: %s" % key)
        return self.request(path, pylastica.request.Request.PUT, args, query)

    def exists(self):
        """
        Checks if the given index exists
        @return:
        @rtype: bool
        """
        response = self._client.request('_cluster/state', pylastica.request.Request.GET, query={'filter_indices': self.name})
        return self.name in response.data['metadata']['indices']

    def create_search(self, query=None, options=None):
        """

        @param query:
        @type query: pylastica.query.Query
        @param options:
        @type options:  dict
        @return:
        @rtype: pylastica.search.Search
        """
        search = pylastica.search.Search(self.client)
        search.add_index(self)
        search.set_options_and_query(options, query)
        return search

    def search(self, query=None, options=None):
        """
        Searches this index
        @param query: dict of query data or Query object
        @type query: string or dict or pylastica.query.Query or pylastica.query.AbstractQuery or pylastica.suggest.Suggest
        @param options: optional dict of options or integer result limit
        @type options: dict or int
        @return:
        @rtype: pylastica.resultset.ResultSet
        """
        return self.create_search(query, options).search()

    def count(self, query=None):
        """
        Counts results for a query. If no query is set, a MatchAll query is used.
        @param query: dict with all query data or a Query object
        @type query: dict or pylastica.query.Query
        @return: number of docs matching the query
        @rtype: int
        """
        return self.create_search(query).count()

    def open(self):
        """
        Open the index
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request('_open', pylastica.request.Request.POST)

    def close(self):
        """
        Close the index
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request('_close', pylastica.request.Request.POST)

    @property
    def name(self):
        """
        Returns the name of the index
        @return:
        @rtype: str
        """
        return self._name

    @property
    def client(self):
        """
        Returns the client object
        @return:
        @rtype: pylastica.client.Client
        """
        return self._client

    def add_alias(self, name, replace=False):
        """
        Adds an alias to the current index
        @param name: alias name
        @type name: str
        @param replace: if True, existing alias of the same name will be replaced
        @type replace: bool
        @return:
        @rtype: pylastica.response.Response
        """
        data = {'actions': []}
        if replace:
            status = pylastica.Status(self.client)
            for index in status.get_indices_with_alias(name):
                data['actions'].append({
                    'remove':{
                        'index': index.name,
                        'alias': name
                    }
                })
        data['actions'].append({
            'add':{
                'index': self.name,
                'alias': name
            }
        })
        return self.client.request('_aliases', pylastica.request.Request.POST, data)

    def remove_alias(self, name):
        """
        Removes an alias of the given name pointing to the current index
        @param name:
        @type name: str
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request('_alias/%s' % name, pylastica.request.Request.DELETE)

    def clear_cache(self):
        """
        Clears the cache of this index
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request('_cache/clear', pylastica.request.Request.POST)

    def flush(self):
        """
        Flushes the index to storage
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request('_flush', pylastica.request.Request.POST)

    def set_settings(self, data):
        """
        Can be used to change settings during runtime.
        @param data:
        @type data: dict
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request('_settings', pylastica.request.Request.PUT, data)

    def request(self, path, method, data=None, query=None):
        """
        Make a call to the ES server based on the current index
        @param path: url
        @type path: str
        @param method: REST method
        @type method: str
        @param data: optional request data
        @type data: dict
        @param query: optional query parameters
        @type query: dict
        @return:
        @rtype: pylastica.response.Response
        """
        path = '/'.join([self.name, path])
        return self.client.request(path, method, data, query)



__author__ = 'Joe Linn'

#import pylastica
import pylastica.param
import pylastica.transport

class Connection(pylastica.param.Param):
    DEFAULT_PORT = 9200
    DEFAULT_HOST = 'localhost'
    DEFAULT_TRANSPORT = 'Http'
    TIMEOUT = 300

    def __init__(self, params=None):
        """

        @param params: optional connection params (host, port, transport, timeout, etc.)
        @type params: dict
        """
        super(Connection, self).__init__()
        self.params = params
        self.enabled = True
        if not self.has_param('config'):
            self.set_param('config', {})

    @property
    def port(self):
        """

        @return:
        @rtype: int
        """
        return self.get_param('port') if self.has_param('port') and self.get_param('port') is not None else self.DEFAULT_PORT

    @port.setter
    def port(self, port):
        """

        @param port:
        @type port: int
        """
        self.set_param('port', int(port))

    @property
    def host(self):
        """

        @return:
        @rtype: str
        """
        return self.get_param('host') if self.has_param('host') else self.DEFAULT_HOST

    @host.setter
    def host(self, host):
        """

        @param host:
        @type host: str
        """
        self.set_param('host', host)

    @property
    def transport(self):
        """

        @return:
        @rtype: str or dict
        """
        return self.get_param('transport') if self.has_param('transport') and self.get_param('transport') is not None else self.DEFAULT_TRANSPORT

    @transport.setter
    def transport(self, transport):
        """
        @param transport:
        @type transport: str
        """
        self.set_param('transport', transport)

    @property
    def path(self):
        """

        @return:
        @rtype: str
        """
        return self.get_param('path') if self.has_param('path') and self.get_param('path') is not None else ''

    @path.setter
    def path(self, path):
        """

        @param path:
        @type path: str
        """
        self.set_param('path', path)

    @property
    def timeout(self):
        """

        @return: connection timeout in seconds
        @rtype: int
        """
        return int(self.get_param('timeout')) if self.has_param('timeout') and self.get_param('timeout') is not None else self.TIMEOUT

    @timeout.setter
    def timeout(self, timeout):
        """

        @param timeout:
        @type timeout: int
        """
        self.set_param('timeout', int(timeout))

    @property
    def enabled(self):
        """

        @return:
        @rtype: bool
        """
        return bool(self.get_param('enabled'))

    @enabled.setter
    def enabled(self, enabled):
        """

        @param enabled:
        @type enabled: bool
        """
        self.set_param('enabled', bool(enabled))

    def is_enabled(self):
        """

        @return:
        @rtype: bool
        """
        return self.enabled

    def get_transport_object(self):
        """
        Returns an instance of the transport type
        @return:
        @rtype: pylastica.transport.AbstractTransport
        """
        return pylastica.transport.AbstractTransport.create(self.transport, self)

    def is_persistent(self):
        """

        @return: True if connection is persistent (default)
        @rtype: bool
        """
        return bool(self.get_param('persistent')) if self.has_param('persistent') else True

    def set_config(self, config):
        """

        @param config:
        @type config: dict
        @return:
        @rtype: self
        """
        return self.set_param('config', config)

    def add_config(self, key, value):
        """

        @param key:
        @type key: str
        @param value:
        @type value: mixexd
        @return:
        @rtype: self
        """
        self._params['config'][key] = value
        return self

    def has_config(self, key):
        """

        @param key:
        @type key: str
        @return:
        @rtype: bool
        """
        return key in self.get_config()

    def get_config(self, key=None):
        """
        Returns a specific config key or the whole config dict
        @param key:
        @type key: str
        @return: config value(s)
        @rtype: dict or str
        """
        config = self.get_param('config')
        if key is None:
            return config
        if key not in config:
            raise pylastica.exception.InvalidException("Config key %s is not set." % key)
        return config[key]

    @classmethod
    def create(cls, params=None):
        """

        @param cls:
        @type cls:
        @param params: params to create a connection
        @type params: dict or pylastica.connection.Connection
        @return:
        @rtype: cls
        """
        connection = None
        if isinstance(params, Connection):
            connection = params
        elif isinstance(params, dict):
            connection = cls(params)
        else:
            raise pylastica.exception.InvalidException('Invalid connection params: %r' % params)
        return connection



__author__ = 'Joe Linn'

from .abstractfacet import AbstractFacet
import pylastica.script


class TermsStats(AbstractFacet):
    def set_key_field(self, key_field):
        """
        Sets the key field for the query
        @param key_field:
        @type key_field: str
        @return:
        @rtype: self
        """
        return self.set_param('key_field', key_field)

    def set_value_script(self, script):
        """
        Set a script to calculate the statistical information on a per term basis
        @param script:
        @type script: pylastica.script.Script
        @return:
        @rtype: self
        """
        if isinstance(script, pylastica.script.Script):
            raise TypeError("script must be of type Script: %r" % script)
        return self.set_param('value_script', script.script).set_param('params', script.params)

    def set_value_field(self, field):
        """
        Set the field on which to compute stats
        @param field:
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param('value_field', field)

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        self._set_facet_param('terms_stats', self._params)
        return super(TermsStats, self).to_dict()





__author__ = 'Joe Linn'

import pylastica.param


class Script(pylastica.param.Param):
    LANG_MVEL = 'mvel'
    LANG_JS = 'js'
    LANG_GROOVY = 'groovy'
    LANG_PYTHON = 'python'
    LANG_NATIVE = 'native'

    def __init__(self, script, params=None, lang=None, doc_id=None):
        """
        @param script: script
        @type script: str
        @param params: script params
        @type params: dict
        @param lang: script language
        @type lang: str
        """
        super(Script, self).__init__()
        self._script = script
        self._lang = lang
        self._upsert = None
        if params is not None:
            assert isinstance(params, dict), "params must be of type dict: %r" % params
            self.params = params
        if doc_id is not None:
            self.doc_id = doc_id

    @property
    def upsert(self):
        """

        @return:
        @rtype: pylastica.document.Document
        """
        return self._upsert

    @upsert.setter
    def upsert(self, data):
        """

        @param data:
        @type data: dict or pylastica.document.Document
        """
        document = pylastica.document.Document.create(data)
        self._upsert = document

    def has_upsert(self):
        """

        @return:
        @rtype: bool
        """
        return self._upsert is not None

    @property
    def doc_id(self):
        """

        @return:
        @rtype: str
        """
        return self.get_param('_id') if self.has_param('_id') else None

    @doc_id.setter
    def doc_id(self, doc_id):
        """

        @param doc_id:
        @type doc_id: str
        """
        self.set_param('_id', doc_id)

    def has_doc_id(self):
        """

        @return:
        @rtype: bool
        """
        return self.has_param('_id')

    def get_options(self, fields=None, with_underscore=False):
        """

        @param fields: if None, all options will be returned. Field names can be with or without underscores. (_percolate, routing)
        @type fields: list of str
        @param with_underscore: if true, option keys should contain an underscore prefix
        @type with_underscore: bool
        @return:
        @rtype: dict
        """
        def strip_underscore(string):
            if not with_underscore:
                return string.lstrip('_')
            return string

        if fields is not None:
            data = {}
            for field in fields:
                key = '_' + field.lstrip('_')
                if self.has_param(key) and str(self.get_param(key)) != '':
                    data[strip_underscore(key)] = self.get_param(key)
        else:
            data = {strip_underscore(key): value for key, value in self.params.iteritems()}
        return data

    @property
    def lang(self):
        """
        Get this script's set language
        @return:
        @rtype: str
        """
        return self._lang

    @lang.setter
    def lang(self, lang):
        """
        Set the language for this script
        @param lang: see LANG_* class properties
        @type lang: str
        """
        self._lang = lang

    def set_lang(self, lang):
        """
        Set the language for this script
        @param lang: see LANG_* class properties
        @type lang: str
        @return:
        @rtype: self
        """
        self._lang = lang
        return self

    @property
    def script(self):
        """
        Get the script
        @return:
        @rtype: str
        """
        return self._script

    @script.setter
    def script(self, script):
        """
        Set the script
        @param script:
        @type script: str
        """
        self._script = script

    def set_script(self, script):
        """
        Set the script
        @param script:
        @type script: str
        @return:
        @rtype: self
        """
        self._script = script
        return self

    @classmethod
    def create(cls, data):
        """

        @param cls:
        @type cls: Script
        @param data:
        @type data: str or dict or Script
        @return:
        @rtype: self
        """
        if isinstance(data, Script):
            script = data
        elif isinstance(data, dict):
            script = Script._create_from_dict(data)
        elif isinstance(data, str):
            script = cls(data)
        else:
            raise pylastica.exception.InvalidException("Failed to create script. Invalid data passed: %r" % data)
        return script

    @classmethod
    def _create_from_dict(cls, data):
        """

        @param cls:
        @type cls: Script
        @param data:
        @type data: dict
        @return:
        @rtype: Script
        """
        if 'script' not in data:
            raise pylastica.exception.InvalidException("data['script'] is required.")
        script = cls(data['script'])
        if 'lang' in data:
            script.lang = data['lang']
        if 'params' in data:
            assert isinstance(data['params'], dict), "data['params'] must be a dict: %r" % data['params']
            script.params = data['params']
        return script

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        dictionary = {
            'script': self._script
        }
        if len(self._params):
            dictionary['params'] = self._params
        if self._lang is not None:
            dictionary['lang'] = self._lang
        return dictionary



__author__ = 'Joe Linn'

import pylastica.aggregation.histogram as histogram


class DateHistogram(histogram.Histogram):
    def __init__(self, name, field, interval):
        """
        @param name: the name of this aggregation
        @type name: str
        @param field: the field on which to perform this aggregation
        @type field: str
        @param interval: the interval by which documents will be bucketed
        @type interval: str
        """
        super(DateHistogram, self).__init__(name, field, interval)

    def set_interval(self, interval):
        """
        Set the interval by which documents will be bucketed
        @param interval: "1.5h", for example
        @type interval: str
        @return:
        @rtype: DateHistogram
        """
        return self.set_param("interval", interval)

    def set_pre_zone(self, pre_zone):
        """
        Set pre-rounding based on interval
        @param pre_zone:
        @type pre_zone: str
        @return:
        @rtype: DateHistogram
        """
        return self.set_param("pre_zone", pre_zone)

    def set_post_zone(self, post_zone):
        """
        Set post-rounding based on interval
        @param post_zone:
        @type post_zone: str
        @return:
        @rtype: DateHistogram
        """
        return self.set_param("post_zone", post_zone)

    def set_pre_zone_adjust_large_interval(self, adjust):
        """
        Set pre-zone adjustment for larger time intervals (day and above)
        @param adjust:
        @type adjust: str
        @return:
        @rtype: DateHistogram
        """
        return self.set_param("pre_zone_adjust_large_interval", adjust)

    def set_factor(self, factor):
        """
        Adjust for granularity of date data
        @param factor: set to 1000 if date is stored in seconds rather than milliseconds
        @type factor: int
        @return:
        @rtype: DateHistogram
        """
        return self.set_param("factor", factor)

    def set_pre_offset(self, offset):
        """
        Set the offset for pre-rounding
        @param offset: "1d", for example
        @type offset: str
        @return:
        @rtype: DateHistogram
        """
        return self.set_param("pre_offset", offset)

    def set_post_offset(self, offset):
        """
        Set the offset for post-rounding
        @param offset: "1d", for example
        @type offset: str
        @return:
        @rtype: DateHistogram
        """
        return self.set_param("post_offset", offset)

    def set_format(self, format):
        """
        Set the format for returned bucket key_as_string values
        @param format: see http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/search-aggregations-bucket-daterange-aggregation.html#date-format-pattern
        @type format: str
        @return:
        @rtype: DateHistogram
        """
        return self.set_param("format", format)


__author__ = 'Joe Linn'

import pylastica.param
import pylastica.exception
import pylastica.script

class ScriptFields(pylastica.param.Param):
    def __init__(self, scripts=None):
        """
        @param scripts: dict of name: Script
        @type scripts: dict of (str, pylastica.script.Script)
        """
        super(ScriptFields, self).__init__()
        if scripts is not None:
            self.set_scripts(scripts)

    def add_script(self, name, script):
        """
        Add a script
        @param name: name of the script field
        @type name: str
        @param script: the script
        @type script: pylastica.script.Script
        @return:
        @rtype: self
        """
        assert isinstance(name, str), "name must be a string: %r" % name
        assert len(name) > 0, "name must be set"
        return self.set_param(name, script.to_dict())

    def set_scripts(self, scripts):
        """
        Set multiple scripts
        @param scripts: dict of name: Script
        @type scripts: dict of (str, pylastica.script.Script)
        @return:
        @rtype: self
        """
        self._params = {}
        for name, script in scripts.iteritems():
            self.add_script(name, script)
        return self

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        return self._params



__author__ = 'Joe Linn'

from .bool import *
from .booland import *
from .boolnot import *
from .boolor import *
from .exists import *
from .geoboundingbox import *
from .geodistance import *
from .geodistancerange import *
from .geohashcell import *
from .geopolygon import *
from .haschild import *
from .hasparent import *
from .ids import *
from .limit import *
from .matchall import *
from .missing import *
from .nested import *
from .numericrange import *
from .prefix import *
from .query import *
from .range import *
from .script import *
from .term import *
from .terms import *
from .type import *



__author__ = 'Joe Linn'

import pylastica.exception

class HttpException(pylastica.exception.ConnectionException):
    def __init__(self, error, request=None, response=None):
        """

        @param error:
        @type error: str
        @param request:
        @type request: pylastica.request.Request
        @param response:
        @type response: pylastica.response.Response
        """
        self._error = error
        message = self.get_error_message(self.error)
        super(HttpException, self).__init__(message, request, response)

    def get_error_message(self, error):
        """
        Returns the error message corresponding to the error code
        @param error: error code
        @type error: str
        @return: error message
        @rtype: str
        """
        #TODO: translate http error codes
        return error

    @property
    def error(self):
        """

        @return:
        @rtype: str
        """
        return self._error

class ThriftException(pylastica.exception.ConnectionException):
    def __init__(self, thrift_exception, request=None, response=None, *args, **kwargs):
        """

        @param thrift_exception:
        @type thrift_exception: thrift.Thrift.TException
        @param request:
        @type request: pylastica.request.Request
        @param response:
        @type response: pylastica.response.Response
        @param args:
        @type args:
        @param kwargs:
        @type kwargs:
        """
        self._thrift_exception = thrift_exception
        message = self.error_message
        super(ThriftException, self).__init__(message, request, response, *args, **kwargs)

    @property
    def error_message(self):
        """

        @return:
        @rtype: str
        """
        return self.thrift_exception.message

    @property
    def thrift_exception(self):
        """

        @return:
        @rtype: thrift.Thrift.TException
        """
        return self._thrift_exception



__author__ = 'Joe Linn'

from .abstractfacet import AbstractFacet
import pylastica.query.abstract


class Query(AbstractFacet):
    def set_query(self, query):
        """
        Set the query for this facet
        @param query:
        @type query: pylastica.query.abstract.AbstractQuery
        @return:
        @rtype: self
        """
        if not isinstance(query, pylastica.query.abstract.AbstractQuery):
            raise TypeError("query must be an instance of AbstractQuery: %r" % query)
        return self._set_facet_param('query', query.to_dict())



__author__ = 'Joe Linn'

#import pylastica
from .abstract import AbstractQuery

class CustomFiltersScore(AbstractQuery):
    SCORE_MODE_FIRST = 'first'
    SCORE_MODE_MIN = 'min'
    SCORE_MODE_MAX = 'max'
    SCORE_MODE_TOTAL = 'total'
    SCORE_MOD_AVG = 'avg'
    SCORE_MODE_MULTIPLY = 'multiply'

    def __init__(self, query=None):
        """

        @param query:
        @type query: dict or pylastica.query.Query
        """
        super(CustomFiltersScore, self).__init__()
        self.set_query(query)

    def set_query(self, query):
        """
        Set the query
        @param query:
        @type query: pylastica.query.Query or dict
        @return:
        @rtype: self
        """
        query = pylastica.query.Query.create(query)
        data = query.to_dict()
        self.set_param('query', data['query'])
        return self

    def add_filter(self, filter, boost):
        """
        Add a filter with boost
        @param filter:
        @type filter: pylastica.filter.AbstractFilter
        @param boost: boost for the filter
        @type boost: float
        @return:
        @rtype: self
        """
        assert isinstance(filter, pylastica.filter.AbstractFilter), "filter must be a descendant of AbstractFilter: %r" % filter
        filter_param = {
            'filter': filter.to_dict(),
            'boost': boost
        }
        return self.add_param('filters', filter_param)

    def add_filter_script(self, filter, script):
        """
        Add a filter with a script to calculate the score.
        @param filter: filter object
        @type filter: pylastica.filter.AbstractFilter
        @param script:
        @type script: pylastica.script.Script or str or dict
        @return:
        @rtype: self
        """
        assert isinstance(filter, pylastica.filter.AbstractFilter), "filter must be a descendant of AbstractFilter: %r" % filter
        script = pylastica.script.Script.create(script)
        filter_param = {
            'filter': filter.to_dict(),
            'script': script.script
        }
        return self.add_param('filters', filter_param)

    def set_script_lang(self, lang):
        """
        Set language for scripts in filters
        @param lang:
        @type lang: str
        @return:
        @rtype: self
        """
        return self.set_param('lang', lang)

    def set_script_params(self, params):
        """
        Set parameters for scripts
        @param params:
        @type params: dict
        @return:
        @rtype: self
        """
        return self.set_param('params', params)

    def set_score_mode(self, score_mode):
        """
        Set the score mode.
        @param score_mode: See SCORE_MODE_* class properties for options.
        @type score_mode: str
        @return:
        @rtype: self
        """
        return self.set_param('score_mode', score_mode)



__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter

class Exists(AbstractFilter):
    """
    @see http://www.elasticsearch.org/guide/reference/query-dsl/exists-filter.html
    """
    def __init__(self, field):
        """

        @param field:
        @type field: str
        """
        super(Exists, self).__init__()
        self.set_field(field)

    def set_field(self, field):
        """
        Set field
        @param field:
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param('field', field)



__author__ = 'Joe Linn'

#import pylastica.filter.range
from .range import Range

class NumericRange(Range):
    """
    @see: http://www.elasticsearch.org/guide/reference/query-dsl/numeric-range-filter/
    """
    pass



__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter
import pylastica.query

class HasChild(AbstractFilter):
    def __init__(self, query, doc_type=None):
        """
        @param query: query string or query object
        @type query: str or pylastica.query.Query or pylastica.query.abstract.AbstractQuery
        @param doc_type: child document type
        @type doc_type: str
        """
        super(HasChild, self).__init__()
        self.set_query(query)
        self.set_doc_type(doc_type)

    def set_query(self, query):
        """
        Set the query object
        @param query:
        @type query: str or pylastica.query.Query or pylastica.query.abstract.AbstractQuery
        @return:
        @rtype: self
        """
        query = pylastica.query.Query.create(query)
        data = query.to_dict()
        return self.set_param('query', data['query'])

    def set_doc_type(self, doc_type):
        """
        Set the type of the child document
        @param doc_type: child document type
        @type doc_type: str
        @return:
        @rtype: self
        """
        return self.set_param('type', doc_type)

    def set_scope(self, scope):
        """
        Set the scope
        @param scope:
        @type scope: str
        @return:
        @rtype: self
        """
        return self.set_param('_scope', scope)



__author__ = 'Joe Linn'

from .abstract import AbstractQuery

class MultiMatch(AbstractQuery):
    def set_query(self, query=None):
        """
        Set the query string
        @param query:
        @type query: str
        @return:
        @rtype: self
        """
        if query is None:
            query = ''
        return self.set_param('query', query)

    def set_fields(self, fields=None):
        """
        Set the fields to be used in the query
        @param fields:
        @type fields: list of str
        @return:
        @rtype: self
        """
        if fields is None:
            fields = []
        return self.set_param('fields', fields)

    def set_use_dis_max(self, use_dis_max=True):
        """
        Sets use dis max, indicating to either create a dis_max query or a bool query.
        @param use_dis_max:
        @type use_dis_max: bool
        @return:
        @rtype: self
        """
        return self.set_param('use_dis_max', bool(use_dis_max))

    def set_tie_breaker(self, tie_breaker=0.0):
        """
        Set the tie breaker multiplier value to balance scores between lower and higher scoring fields.
        @param tie_breaker:
        @type tie_breaker: float
        @return:
        @rtype: self
        """
        return self.set_param('tie_breaker', float(tie_breaker))



__author__ = 'Joe Linn'

from .abstract import AbstractQuery
import pylastica.exception

class Bool(AbstractQuery):
    def add_should(self, args):
        """
        Add should part to query
        @param args:
        @type args: pylastica.query.AbstractQuery or dict
        @return:
        @rtype: self
        """
        return self._add_query('should', args)

    def add_must(self, args):
        """
        Add must part to the query
        @param args:
        @type args: pylastica.query.AbstractQuery or dict
        @return:
        @rtype: self
        """
        return self._add_query('must', args)

    def add_must_not(self, args):
        """
        Add must not part
        @param args:
        @type args: pylastica.query.AbstractQuery or dict
        @return:
        @rtype: self
        """
        return self._add_query('must_not', args)

    def _add_query(self, type, args):
        """
        Adds a query to the current object
        @param type: query type
        @type type: str
        @param args: query
        @type args: pylastica.query.AbstractQuery or dict
        @return:
        @rtype: bool
        """
        if isinstance(args, AbstractQuery):
            args = args.to_dict()
        assert isinstance(args, dict), "Invalid parameter. Must be array or instance of pylastica.query.AbstractQuery: %r" % args
        return self.add_param(type, args)

    def set_boost(self, boost):
        """
        Set the boost value of this query
        @param boost:
        @type boost: float
        @return:
        @rtype: self
        """
        return self.set_param('boost', boost)

    def set_minimum_number_should_match(self, minimum):
        """
        Set the minimum number of docs which should match this query
        @param minimum:
        @type minimum: int
        @return:
        @rtype: self
        """
        return self.set_param('minimum_number_should_match', minimum)



__author__ = 'Joe Linn'

import pylastica.aggregation.abstractaggregation as abstract


class Range(abstract.SimpleAggregation):
    def add_range(self, from_value=None, to_value=None):
        """
        Add a range for this aggregation
        @param from_value: low end of this range, exclusive (greater than)
        @type from_value: int or float
        @param to_value: high end of this range, exclusive (less than)
        @type to_value: int or float
        @return:
        @rtype: Range
        """
        if from_value is None and to_value is None:
            raise ValueError("Either from_value or to_value must be set. Both cannot be None.")
        range = {}
        if from_value is not None:
            range['from'] = from_value
        if to_value is not None:
            range['to'] = to_value
        return self.add_param('ranges', range)

    def set_keyed_response(self, keyed=True):
        """
        If set to True, a unique string key will be associated with each bucket and ranges will be returned as a
        dictionary rather than a list.
        @param keyed:
        @type keyed: bool
        @return:
        @rtype: Range
        """
        return self.set_param('keyed', keyed)


__author__ = 'Joe Linn'

import pylastica.exception


class ActionException(pylastica.exception.BulkException):
    def __init__(self, response, *args, **kwargs):
        """

        @param response:
        @type response: pylastica.bulk.response.Response
        @param args:
        @type args:
        @param kwargs:
        @type kwargs:
        @return:
        @rtype:
        """
        self._response = response
        super(ActionException, self).__init__(self.get_error_message(response), *args, **kwargs)

    @property
    def action(self):
        """

        @return:
        @rtype: pylastica.bulk.action.Action
        """
        return self.response.action

    @property
    def response(self):
        """

        @return:
        @rtype: pylastica.bulk.response.Response
        """
        return self._response

    def get_error_message(self, response):
        """

        @param response:
        @type response: pylastica.bulk.response.Response
        @return:
        @rtype: str
        """
        error = response.error
        op_type = response.op_type
        data = response.data
        path = ''
        if '_index' in data:
            path += '/' + data['_index']
        if '_type' in data:
            path += '/' + data['_type']
        if '_id' in data:
            path += '/' + data['_id']
        return "%s: %s caused %s" % (op_type, path, error)



#
# Autogenerated by Thrift
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
from __future__ import absolute_import

from thrift.Thrift import *
from .ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
    from thrift.protocol import fastbinary
except ImportError:
    fastbinary = None


class Iface:
    def execute(self, request):
        """
        Parameters:
         - request
        """
        pass


class Client(Iface):
    def __init__(self, iprot, oprot=None):
        self._iprot = self._oprot = iprot
        if oprot != None:
            self._oprot = oprot
        self._seqid = 0

    def execute(self, request):
        """
        Parameters:
         - request
        """
        self.send_execute(request)
        return self.recv_execute()

    def send_execute(self, request):
        self._oprot.writeMessageBegin('execute', TMessageType.CALL, self._seqid)
        args = execute_args()
        args.request = request
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_execute(self):
        (fname, mtype, rseqid) = self._iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(self._iprot)
            self._iprot.readMessageEnd()
            raise x
        result = execute_result()
        result.read(self._iprot)
        self._iprot.readMessageEnd()
        if result.success != None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "execute failed: unknown result");


class Processor(Iface, TProcessor):
    def __init__(self, handler):
        self._handler = handler
        self._processMap = {}
        self._processMap["execute"] = Processor.process_execute

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_execute(self, seqid, iprot, oprot):
        args = execute_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = execute_result()
        result.success = self._handler.execute(args.request)
        oprot.writeMessageBegin("execute", TMessageType.REPLY, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()


# HELPER FUNCTIONS AND STRUCTURES

class execute_args:
    """
    Attributes:
     - request
    """

    thrift_spec = (
        None, # 0
        (1, TType.STRUCT, 'request', (RestRequest, RestRequest.thrift_spec), None, ), # 1
    )

    def __init__(self, request=None):
        self.request = request

    def read(self, iprot):
        if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
            fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.request = RestRequest()
                    self.request.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
            oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('execute_args')
        if self.request != None:
            oprot.writeFieldBegin('request', TType.STRUCT, 1)
            self.request.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()
        def validate(self):
            if self.request is None:
                raise TProtocol.TProtocolException(message='Required field request is unset!')
            return


    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.iteritems()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)

class execute_result:
    """
    Attributes:
     - success
    """

    thrift_spec = (
        (0, TType.STRUCT, 'success', (RestResponse, RestResponse.thrift_spec), None, ), # 0
    )

    def __init__(self, success=None):
        self.success = success

    def read(self, iprot):
        if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
            fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = RestResponse()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
            oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('execute_result')
        if self.success != None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()
        def validate(self):
            return


    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.iteritems()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


__author__ = 'Joe Linn'

import urllib
#import pylastica
from .abstracttype import *
from .mapping import *

class DocType(pylastica.searchable.Searchable):
    def __init__(self, index, name):
        """
        @param index: index object
        @type index: pylastica.index.Index
        @param name: document type name
        @type name: str
        """
        assert isinstance(index, pylastica.index.Index), "index must be of type Index: %r" % index
        self._index = index
        self._name = name
        self._serializer = None

    def add_document(self, doc):
        """
        Adds the given document to the index
        @param doc:
        @type doc: pylastica.document.Document
        @return:
        @rtype: pylastica.response.Response
        """
        assert isinstance(doc, pylastica.document.Document), "doc must be of type Document: %r" % doc
        path = urllib.quote_plus(str(doc.doc_id))
        request_type = pylastica.request.Request.PUT
        if path is None or path == '':
            #no doc id has been given; use post so that an id is automatically created
            request_type = pylastica.request.Request.POST
        options = doc.get_options([
            'version',
            'version_type',
            'routing',
            'percolate',
            'parent',
            'ttl',
            'timestamp',
            'op_type',
            'consistency',
            'replication',
            'refresh',
            'timeout'
        ])
        response = self.request(path, request_type, doc.data, options)
        data = response.data
        if (doc.auto_populate or self.index.client.get_config_value(['document', 'autoPopulate'], False)) and response.is_ok():
            if doc.has_id():
                if '_id' in data:
                    doc.doc_id = data['_id']
            if '_version' in data:
                doc.version = data['_version']
        return response

    def update_document(self, data):
        """
        Update a document using an update script.
        @param doc: document with update data
        @type doc: pylastica.document.Document or pylastica.script.Script
        @return:
        @rtype: pylastica.response.Response
        """
        if not isinstance(data, pylastica.document.Document) and not isinstance(data, pylastica.script.Script):
            raise TypeError("data must be an instance of Document or Script: %r" % data)
        if not data.has_id():
            raise pylastica.exception.InvalidException("Document id is not set.")
        return self.index.client.update_document(data.doc_id, data, self.index.name, self.name)

    def add_documents(self, docs):
        """
        Use _bulk to send multiple documents to the server
        @param docs: list of doucments
        @type docs: list of pylastica.document.Document
        @return:
        @rtype: pylastica.bulk.ResponseSet
        """
        for doc in docs:
            assert isinstance(doc, pylastica.document.Document), "All items in list docs must be of type Document: %r" % doc
            doc.doc_type = self.name
        return self.index.add_documents(docs)

    def get_document(self, doc_id, options=None):
        """
        Retrieve a specific document from the index
        @param doc_id: document id
        @type doc_id: str
        @param options: options for the get request
        @type options: dict
        @return:
        @rtype: pylastica.document.Document
        """
        path = urllib.quote_plus(str(doc_id))
        try:
            result = self.request(path, pylastica.request.Request.GET, query=options).data
        except pylastica.exception.ResponseException:
            raise pylastica.exception.NotFoundException("Document with id %s not found." % doc_id)
        if result['found'] is None or result['found'] == '' or not result['found']:
            raise pylastica.exception.NotFoundException("Document with id %s not found." % doc_id)
        data = result['_source'] if '_source' in result else {}
        document = pylastica.document.Document(doc_id, data, self.name, self.index)
        document.version = result['_version']
        return document

    def create_document(self, doc_id=None, data=None):
        """

        @param doc_id:
        @type doc_id: str or int
        @param data:
        @type data: dict or str
        @return:
        @rtype: pylastica.document.Document
        """
        return pylastica.document.Document(doc_id, data, doc_type=self, index=self.index)

    @property
    def name(self):
        """

        @return: type name
        @rtype: str
        """
        return self._name

    def set_mapping(self, mapping):
        """
        Sets the value type mapping
        @param mapping:
        @type mapping: pylastica.doc_type.Mapping or dict
        @return:
        @rtype: pylastica.response.Response
        """
        mapping = pylastica.doc_type.Mapping.create(mapping)
        mapping.doc_type = self
        return mapping.send()

    @property
    def mapping(self):
        """
        Returns the current mapping for this type
        @return:
        @rtype: dict
        """
        return self.request('_mapping', pylastica.request.Request.GET).data

    @mapping.setter
    def mapping(self, mapping):
        """
        Sets the value type mapping.
        @param mapping:
        @type mapping: pylastica.type.Mapping or dict
        @return:
        @rtype: pylastica.response.Response
        """
        self.set_mapping(mapping)

    def create_search(self, query=None, options=None):
        """

        @param query:
        @type query: pylastica.query.Query
        @param options:
        @type options:  dict
        @return:
        @rtype: pylastica.search.Search
        """
        search = pylastica.search.Search(self.index.client)
        return search.add_index(self.index).add_type(self).set_options_and_query(options, query)

    def search(self, query=None, options=None):
        """
        Searches results for a query
        @param query: dict with all query data or a Query object
        @type query: str or dict or pylastica.query.Query or pylastica.query.AbstractQuery
        @param options:
        @type options: dict
        @return: result set with all results
        @rtype: pylastica.resultset.ResultSet
        """
        return self.create_search(query, options).search()

    def count(self, query=None):
        """
        Counts results for a query. If no query is set, a MatchAll query is used.
        @param query: dict with all query data or a Query object
        @type query: dict or pylastica.query.Query or str
        @return: number of docs matching the query
        @rtype: int
        """
        return self.create_search(query).count()

    @property
    def index(self):
        """
        Returns the index object associated with this type
        @return:
        @rtype: pylastica.index.Index
        """
        return self._index

    def delete_document(self, document):
        """
        Delete the given document from the current index
        @param document:
        @type document: pylastica.document.Document
        @return:
        @rtype: pylastica.response.Response
        """
        assert isinstance(document, pylastica.document.Document), "document must be a Document object: %r" % document
        options = document.get_options([
            'version',
            'version_type',
            'routing',
            'parent',
            'replication',
            'consistency',
            'refresh',
            'timeout'
        ])
        return self.delete_by_id(document.doc_id, options)

    def delete_by_id(self, doc_id, options=None):
        """
        Deletes document by its id
        @param doc_id:
        @type doc_id: str
        @param options: query options
        @type options: dict
        @return:
        @rtype: pylastica.response.Response
        """
        if doc_id is None or doc_id == '':
            raise ValueError()
        doc_id = urllib.urlencode(doc_id)
        response = self.request(doc_id, pylastica.request.Request.DELETE, query=options)
        response_data = response.data
        if 'found' in response_data and not response_data['found']:
            raise pylastica.exception.NotFoundException("Document with id %s could not be found." % doc_id)
        return response

    def delete_by_query(self, query):
        """
        Delete documents based on the given query
        @param query:
        @type query: str or pylastica.query.Query
        @return:
        @rtype: pylastica.response.Response
        """
        query = pylastica.query.Query.create(query)
        return self.request('_query', pylastica.request.Request.DELETE, query.query)

    def delete(self):
        """
        Delete the current type from the index
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request('', pylastica.request.Request.DELETE)

    def more_like_this(self, doc, params=None, query=None):
        """
        Perform a more like this query based on the given document
        @param doc:
        @type doc: pylastica.document.Document
        @param params: additional query params
        @type params: dict
        @param query: optional query to filter the more like this results
        @type query: str or dict or pylastica.query.Query
        @return:
        @rtype: pylastica.resultset.ResultSet
        """
        path = "%s/_mlt" % doc.doc_id
        query = pylastica.query.Query.create(query)
        response = self.request(path, pylastica.request.Request.GET, query.to_dict(), params)
        return pylastica.resultset.ResultSet(response, query)

    def request(self, path, method, data=None, query=None):
        """
        Make a call to the ES server based on this doc type
        @param path: path to call (url)
        @type path: str
        @param method: REST method
        @type method: str
        @param data: arguments as a dict
        @type data: dict
        @param query: query parameters
        @type query: dict
        @return:
        @rtype: pylastica.response.Response
        """
        path = '/'.join([self.name, path])
        return self.index.request(path, method, data, query)

    def set_serializer(self, serializer):
        """
        Set the serializer callable used in add_object
        @param serializer:
        @type serializer:
        @return:
        @rtype: self
        """
        self._serializer = serializer
        return self



__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter

class Term(AbstractFilter):
    def __init__(self, field=None, value=None):
        """

        @param field: document field
        @type field: str
        @param value:
        @type value: str or dict
        """
        super(Term, self).__init__()
        if field is not None:
            self.set_raw_term({field: value})

    def set_raw_term(self, term):
        """
        Set / overwrite key and term
        @param term: {field: value}
        @type term: dict
        @return:
        @rtype: self
        """
        self.params = term
        return self

    def set_term(self, field, value):
        """
        Set the term for the filter
        @param field: document field
        @type field: str
        @param value: value for the filter
        @type value: str
        @return:
        @rtype: self
        """
        return self.set_raw_term({field: value})



__author__ = 'Joe Linn'

import pylastica.index

class Status(object):
    def __init__(self, client):
        """

        @param client:
        @type client: pylastica.client.Client
        """
        super(Status, self).__init__()
        self._client = client
        self._data = None
        self._response = None
        self.refresh()

    @property
    def data(self):
        """
        Get status data
        @return:
        @rtype: dict
        """
        return self._data

    @property
    def index_statuses(self):
        """
        Get status objects of all indices
        @return:
        @rtype: list of pylastica.index.status.Status
        """
        return [pylastica.index.Status(pylastica.index.Index(self._client, name))for name in self.index_names]

    @property
    def index_names(self):
        """
        Get the names of existing indices
        @return:
        @rtype: list of str
        """
        return [name for name in self._data['indices']] if len(self._data['indices']) else []

    def index_exists(self, name):
        """
        Check if the given index exists
        @param name: index name
        @type name: str
        @return:
        @rtype: bool
        """
        return name in self.index_names

    def alias_exists(self, alias):
        """
        Determine if the given alias exists
        @param alias:
        @type alias: str
        @return:
        @rtype: bool
        """
        response = self._client.request('/_alias/%s' % alias)
        if response.has_error() and ('AliasMissingException' in response.error or response.status == 404):
            return False
        return True

    def get_indices_with_alias(self, alias):
        """
        Get a list of all indices which share the given alias
        @param alias:
        @type alias: str
        @return:
        @rtype: list of pylastica.index.Index
        @raise: pylastica.exception.AliasMissingException if the requested alias does not exist
        """
        response = self._client.request('_alias/%s' % alias)
        if response.has_error() and ('AliasMissingException' in response.error or response.status == 404):
            #the requested alias does not exist
            return []
        indices = response.data
        return [pylastica.index.Index(self._client, name) for name in indices]

    @property
    def response(self):
        """
        Return the response object
        @return:
        @rtype: pylastica.response.Response
        """
        return self._response

    @property
    def shards(self):
        """
        Return shards info
        @return:
        @rtype: dict
        """
        return self._data['shards']

    def refresh(self):
        """
        Refresh the status object
        @return:
        @rtype: void
        """
        self._response = self._client.request('_status')
        self._data = self.response.data

    @property
    def server_status(self):
        """
        Get server status
        @return:
        @rtype: dict
        """
        return self._client.request('').data



__author__ = 'Joe Linn'

from .abstract import AbstractQuery


class QueryString(AbstractQuery):
    def __init__(self, query_string=''):
        """
        @param query_string: optional query string
        @type query_string: str
        """
        super(QueryString, self).__init__()
        self.set_query(query_string)

    def set_query(self, query):
        """
        Sets a new query string for the object
        @param query:
        @type query: str
        @return:
        @rtype: self
        """
        assert isinstance(query, str), "query must be a str: %r" % query
        self._query_string = query
        return self

    def set_default_field(self, field):
        """
        Set the default field
        @param field: field name
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param('default_field', field)

    def set_default_operator(self, operator):
        """
        Set the default operator
        @param operator: AND or OR
        @type operator: str
        @return:
        @rtype: self
        """
        return self.set_param('default_operator', operator)

    def set_analyzeer(self, analyzer):
        """
        Sets the analyzer for this query
        @param analyzer:
        @type analyzer: str
        @return:
        @rtype: self
        """
        return self.set_param('analyzer', analyzer)

    def set_allow_leading_wildcard(self, allow=True):
        """
        Sets the parameter to allow * and ? as first characters
        @param allow:
        @type allow: bool
        @return:
        @rtype: self
        """
        return self.set_param('allow_leading_wildcard', bool(allow))

    def set_lowercase_expanded_terms(self, lowercase=True):
        """
        Sets the parameter to auto-lowercase terms of queries
        @param lowercase:
        @type lowercase: bool
        @return:
        @rtype: self
        """
        return self.set_param('lowercase_expanded_terms', bool(lowercase))

    def set_enable_position_increments(self, enabled=True):
        """
        Enable or disable position increments in result queries
        @param enabled:
        @type enabled: bool
        @return:
        @rtype: self
        """
        return self.set_param('enable_position_incremends', bool(enabled))

    def set_fuzzy_prefix_length(self, length=0):
        """
        Set the fuzzy prefix length parameter
        @param length:
        @type length: int
        @return:
        @rtype: self
        """
        return self.set_param('fuzzy_prefix_length', int(length))

    def set_phrase_stop(self, phrase_stop=0):
        """
        Set the phrase stop. If 0, exact phrases are required.
        @param phrase_stop:
        @type phrase_stop: int
        @return:
        @rtype: self
        """
        return self.set_param('phrase_stop', int(phrase_stop))

    def set_boost(self, boost=1.0):
        """
        Set the boost for the query
        @param boost:
        @type boost: float
        @return:
        @rtype: self
        """
        return self.set_param('boost', float(boost))

    def set_analyze_wildcard(self, analyze=True):
        """
        Allows analysis of wildcard terms
        @param analyze:
        @type analyze: bool
        @return:
        @rtype: self
        """
        return self.set_param('analyze_wildcard', bool(analyze))

    def set_auto_generate_phrase_queries(self, auto=True):
        """
        Automatically generate phrase queries
        @param auto:
        @type auto: bool
        @return:
        @rtype: self
        """
        return self.set_param('auto_generate_phrase_queries', bool(auto))

    def set_fields(self, fields):
        """
        Set the fields. If no fields are set, _all is used.
        @param fields:
        @type fields: list of str
        @return:
        @rtype: self
        """
        assert isinstance(fields, list), "fields must be a list: %r" % fields
        return self.set_param('fields', fields)

    def set_use_dis_max(self, value=True):
        """
        Set whether or not to use bool or dis_max queries to internally combine results for multi field search
        @param value:
        @type value: bool
        @return:
        @rtype: self
        """
        return self.set_param('use_dis_max', bool(value))

    def set_tie_breaker(self, tie_breaker=0):
        """
        When using dis_max, the disjunction max tie breaker
        @param tie_breaker:
        @type tie_breaker: int
        @return:
        @rtype: self
        """
        return self.set_param('tie_breaker', int(tie_breaker))

    def set_rewrite(self, rewrite=None):
        """
        Set the rewrite condition
        @param rewrite:
        @type rewrite: str
        @return:
        @rtype: self
        """
        if rewrite is None:
            rewrite = ''
        return self.set_param('rewrite', rewrite)

    def to_dict(self):
        dictionary = {'query': self._query_string}
        dictionary.update(self.params)
        return {
            'query_string': dictionary
        }




__author__ = 'Joe Linn'

from .abstract import AbstractQuery
import pylastica.script

class CustomScore(AbstractQuery):
    def __init__(self, script=None, query=None):
        """

        @param script:
        @type script: str or dict or pylastica.script.Script
        @param query:
        @type query: str or pylastica.query.AbstractQuery
        """
        super(CustomScore, self).__init__()
        if script is not None:
            self.set_script(script)
        self.set_query(query)

    def set_query(self, query):
        """
        Set the query object
        @param query:
        @type query: str or pylastica.query.Query or pylastica.query.AbstractQuery
        @return:
        @rtype: self
        """
        query = pylastica.query.Query.create(query)
        data = query.to_dict()
        return self.set_param('query', data['query'])

    def set_script(self, script):
        """
        Set the script
        @param script:
        @type script: str or pylastica.script.Script or dict
        @return:
        @rtype: self
        """
        script = pylastica.script.Script.create(script)
        for param, value in script.to_dict().iteritems():
            self.set_param(param, value)
        return self

    def add_params(self, params):
        """
        Add params for the script
        @param params:
        @type params: dict
        @return:
        @rtype: self
        """
        return self.set_param('params', params)



__author__ = 'Joe Linn'

from .abstractfacet import AbstractFacet
import pylastica.filter.abstractfilter


class Filter(AbstractFacet):
    def set_filter(self, filter_object):
        """
        Set the filter for the facet
        @param filter_object:
        @type filter_object: pylastica.filter.abstractfilter.AbstractFilter
        @return:
        @rtype: self
        """
        if not isinstance(filter_object, pylastica.filter.abstractfilter.AbstractFilter):
            raise TypeError("filter_object must be an instance of an implementation of AbstractFilter: %r" % filter_object)
        return self._set_facet_param('filter', filter_object.to_dict())




__author__ = 'Joe Linn'

from .abstract import *


class ConstantScore(AbstractQuery):
    def __init__(self, filter=None):
        """

        @param filter:
        @type filter: dict or pylastica.filter.AbstractFilter
        """
        super(ConstantScore, self).__init__()
        if filter is not None:
            self.set_filter(filter)

    def set_filter(self, filter):
        """
        Set the filter for this query
        @param filter:
        @type filter: dict or pylastica.filter.AbstractFilter
        @return:
        @rtype: self
        """
        if isinstance(filter, pylastica.filter.AbstractFilter):
            filter = filter.to_dict()
        return self.set_param('filter', filter)

    def set_boost(self, boost):
        """
        Set boost
        @param boost:
        @type boost: float
        @return:
        @rtype: self
        """
        return self.set_param('boost', boost)



__author__ = 'Joe Linn'

import json
import logging


class Log(object):
    def __init__(self, log=''):
        """
        @param log: file name for logging
        @type log: str
        """
        super(Log, self).__init__()
        self._log = True #log path or True if enabled
        self._last_message = ''
        self.set_log(log)

    def log(self, level, message, context=None):
        """

        @param level: log level
        @type level: str
        @param message: message to log
        @type message: str
        @param context:
        @type context: dict
        @return:
        @rtype: void
        """
        if context is None:
            context = {}
        context['error_message'] = message
        self._last_message = json.dumps(context)
        logging.log(level, message)

    def set_log(self, log):
        """
        Enable / disable logging or set log path
        @param log: enables logging or sets the log path
        @type log: bool or str
        @return:
        @rtype: self
        """
        self._log = log
        if isinstance(log, str):
            #set log file
            logging.basicConfig(filename=log)
        return self

    @property
    def last_message(self):
        """

        @return:
        @rtype: str
        """
        return self._last_message



__author__ = 'Joe Linn'

import pylastica.exception


class ResponseException(pylastica.exception.BulkException):
    def __init__(self, response_set, *args, **kwargs):
        """

        @param response_set:
        @type response_set: pylastica.bulk.responseset.ResponseSet
        @param args:
        @type args:
        @param kwargs:
        @type kwargs:
        @return:
        @rtype:
        """
        assert isinstance(response_set, pylastica.bulk.responseset.ResponseSet), "response_set must be an instance of ResponseSet: %r" % response_set
        self._response_set = None
        self._action_exceptions = []
        self._init(response_set)
        message = "Error in one or more bulk request actions:\n\n"
        message += self.action_exceptions_as_string
        super(ResponseException, self).__init__(message, *args, **kwargs)

    def _init(self, response_set):
        """

        @param response_set:
        @type response_set: pylastica.bulk.responseset.ResponseSet
        @return:
        @rtype: void
        """
        self._response_set = response_set
        for bulk_response in response_set.bulk_responses:
            if bulk_response.has_error():
                self._action_exceptions.append(pylastica.exception.bulk.response.ActionException(bulk_response))

    @property
    def response_set(self):
        """

        @return:
        @rtype: pylastica.bulk.responseset.ResponseSet
        """
        return self._response_set

    @property
    def failures(self):
        """

        @return:
        @rtype: list of str
        """
        return [str(action_exception) for action_exception in self.action_exceptions]

    @property
    def action_exceptions(self):
        """

        @return:
        @rtype: list of pylastica.exception.bulk.response.ActionException
        """
        return self._action_exceptions

    @property
    def action_exceptions_as_string(self):
        """

        @return:
        @rtype: str
        """
        message = ''
        for action_exception in self.action_exceptions:
            message += str(action_exception) + "\n"
        return message

from . import response



__author__ = 'Joe Linn'

from .abstractfilter import AbstractGeoDistance


class GeohashCell(AbstractGeoDistance):
    def __init__(self, key, location, precision=-1, neighbors=False):
        """
        @param key: The field on which to filter
        @type key: str
        @param location: location as dict or geohash {'lat':40.3, 'lon':45.2}
        @type location: str or dict
        @param precision: Optional. Integer length of geohash prefix or distance (3, or "50m")
        @type precision: str or int
        @param neighbors: If true, filters cells next to the given cell. Defaults to False.
        @type neighbors: bool
        """
        super(GeohashCell, self).__init__(key, location)
        self.set_precision(precision)
        self.set_neighbors(neighbors)

    def set_precision(self, precision):
        """
        Set the precision for this filter
        @param precision: Integer length of geohash prefix or distance (3, or "50m")
        @type precision: str or int
        @return:
        @rtype: self
        """
        return self.set_param('precision', precision)

    def set_neighbors(self, neighbors):
        """
        Set the neighbors option for this filter
        @param neighbors: If true, filters cells next to the given cell.
        @type neighbors: bool
        @return:
        @rtype: self
        """
        return self.set_param('neighbors', bool(neighbors))


__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter

class Range(AbstractFilter):
    def __init__(self, field_name=None, args=None):
        """

        @param field_name: field name
        @type field_name: str
        @param args: field arguments
        @type args: dict
        @see: http://www.elasticsearch.org/guide/reference/query-dsl/range-filter/
        """
        super(Range, self).__init__()
        self._fields = {}
        if field_name is not None:
            self.add_field(field_name, args)

    def add_field(self, field_name, args):
        """
        Add a field with arguments to the range filter
        @param field_name: field name
        @type field_name: str
        @param args: field arguments
        @type args: dict
        @return:
        @rtype: self
        """
        self._fields[field_name] = args
        return self

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        self.params = self._fields
        return super(Range, self).to_dict()




__author__ = 'Joe Linn'

import abc
import pylastica

class AbstractType(pylastica.searchable.Searchable):
    __metaclass__ = abc.ABCMeta

    MAX_DOCS_PER_REQUEST = 1000

    _index_name = None  #name of index. Override this in implementing classes.
    _type_name = None   #name of document type. Override this in implementing classes.

    def __init__(self, client=None):
        """
        Reads index and type names from class properties _index_name and _type_name, which must be set in child classes.
        @param client: optional client object
        @type client: pylastica.client.Client
        """
        if client is None:
            client = pylastica.client.Client()
        if self._index_name is None:
            raise pylastica.exception.InvalidException("_index_name must be set!")
        if self._type_name is None:
            raise pylastica.exception.InvalidException("_type_name must be set!")
        self._client = client
        self._index = pylastica.index.Index(client, self._index_name)
        self._doc_type = pylastica.doc_type.DocType(self._index, self._type_name)
        self._mapping = {}
        self._index_params = {}
        self._source = True

    def create(self, recreate=False):
        """
        Creates the index and sets the mapping for this type
        @param recreate: if set to true, will recreate the index if it already exists
        @type recreate: bool
        """
        self.index.create(self._index_params, recreate)
        mapping = pylastica.doc_type.Mapping(self.doc_type)
        mapping.properties = self._mapping
        mapping.source = {'enabled': self._source}
        mapping.send()

    def create_search(self, query=None, options=None):
        """

        @param query:
        @type query: pylastica.query.Query
        @param options:
        @type options:  dict
        @return:
        @rtype: pylastica.search.Search
        """
        return self.doc_type.create_search(query, options)

    def count(self, query=None):
        """
        Counts results for a query. If no query is set, a MatchAll query is used.
        @param query: dict with all query data or a Query object
        @type query: dict or pylastica.query.Query
        @return: number of docs matching the query
        @rtype: int
        """
        return self.doc_type.count(query)

    @property
    def index(self):
        """
        Retrieve the index object
        @return:
        @rtype: pylastica.index.Index
        """
        return self._index

    @property
    def doc_type(self):
        """
        Return the document type object
        @return:
        @rtype: pylastica.doc_type.DocType
        """
        return self._doc_type

    def convert_date(self, date):
        """
        Converts given time to format: 1995-12-31T23:59:59Z
        This is the lucene date format
        @param date:
        @type date: int
        @return:
        @rtype: str
        """
        return pylastica.util.convert_date(date)



__author__ = 'Joe Linn'

import pylastica.aggregation.abstractaggregation as abstract


class Nested(abstract.AbstractAggregation):
    def __init__(self, name, path):
        """
        @param name: the name of this aggregation
        @type name: str
        @param path: the nested path for this aggregation
        @type path: str
        """
        super(Nested, self).__init__(name)
        self.set_path(path)

    def set_path(self, path):
        """
        Set the nested path for this aggregation
        @param path:
        @type path: str
        @return:
        @rtype: Nested
        """
        return self.set_param("path", path)


__author__ = 'Joe Linn'

import pylastica.param
import pylastica.scriptfields
from .abstract import AbstractQuery
from .matchall import MatchAll
from pylastica.suggest.suggest import Suggest
from pylastica.aggregation.abstractaggregation import AbstractAggregation
from pylastica.suggest.abstract import AbstractSuggestion


class Query(pylastica.param.Param):
    def __init__(self, query=None):
        """

        @param query: optional query object
        @type query: dict or pylastica.query.AbstractQuery or Suggest
        """
        super(Query, self).__init__()
        if isinstance(query, dict):
            self.set_raw_query(query)
        elif isinstance(query, AbstractQuery):
            self.query = query
        elif isinstance(query, Suggest):
            self.set_suggest(query)

    @classmethod
    def create(cls, query):
        """
        Transforms a string or dict into a Query object
        @param cls:
        @type cls:
        @param query:
        @type query: mixed
        @return:
        @rtype: cls
        """
        if isinstance(query, Query):
            return query
        elif isinstance(query, AbstractQuery) or isinstance(query, Suggest):
            return cls(query)
        elif isinstance(query, AbstractSuggestion):
            return cls(Suggest(query))
        elif isinstance(query, pylastica.filter.AbstractFilter):
            new_query = cls()
            new_query.set_filter(query)
            return new_query
        elif query is None or query == '':
            return cls(MatchAll())
        elif isinstance(query, str):
            from .querystring import QueryString
            return cls(QueryString(query))
        raise NotImplementedError()

    def set_raw_query(self, query):
        """
        Sets query as a raw dict. Will overwrite all currently set arguments
        @param query:
        @type query: dict
        @return:
        @rtype: self
        """
        assert isinstance(query, dict), "query must be a dict: %r" % query
        self._params = query
        return self

    @property
    def query(self):
        """
        Get the query dictionary
        @return:
        @rtype: dict
        """
        return self.get_param('query')

    @query.setter
    def query(self, query):
        """
        Set the query
        @param query:
        @type query: AbstractQuery
        """
        assert isinstance(query, AbstractQuery), "query must be instance of AbstractQuery: %r" % query
        self.set_param('query', query.to_dict())

    def set_filter(self, filter):
        """
        Set the filter for this query
        @param filter:
        @type filter: pylastica.filter.AbstractFilter
        @return:
        @rtype: self
        """
        assert isinstance(filter, pylastica.filter.AbstractFilter), "filter must be an instance of an implementation of AbstractFilter: %r" % filter
        return self.set_param('filter', filter.to_dict())

    def set_from(self, start_from):
        """
        Sets the start from which the search results should be returned
        @param start_from:
        @type start_from: int
        @return:
        @rtype: self
        """
        return self.set_param('from', int(start_from))

    def set_sort(self, sort_args):
        """
        Sets sort arguments for the query. Replaces existing values.
        @see: http://www.elasticsearch.org/guide/reference/api/search/sort.html
        @param sort_args:
        @type sort_args: dict
        @return:
        @rtype: self
        """
        return self.set_param('sort', sort_args)

    def add_sort(self, sort):
        """
        Adds a sort param to the query
        @param sort:
        @type sort: mixed
        @return:
        @rtype: self
        """
        return self.add_param('sort', sort)

    def set_highlight(self, highlight_args):
        """
        Sets highlight arguments for the query
        @param highlight_args:
        @type highlight_args: dict
        @return:
        @rtype: self
        """
        return self.set_param('highlight', highlight_args)

    def add_highlight(self, highlight):
        """
        Adds a highlight argument
        @param highlight:
        @type highlight: mixed
        @return:
        @rtype: self
        """
        return self.add_param('highlight', highlight)

    def set_size(self, size=10):
        """
        Set the maximum number of results for this query
        @param size:
        @type size: int
        @return:
        @rtype: self
        """
        return self.set_param('size', int(size))

    def set_explain(self, explain=True):
        """
        Enables explain on the query
        @param explain:
        @type explain: bool
        @return:
        @rtype: self
        """
        return self.set_param('explain', bool(explain))

    def set_version(self, version=True):
        """
        Enables version on the query
        @param version:
        @type version: bool
        @return:
        @rtype: self
        """
        return self.set_param('version', bool(version))

    def set_fields(self, fields):
        """
        Sets the fields to be returned by the query
        @param fields: list of field names
        @type fields: list of str
        @return:
        @rtype: self
        """
        return self.set_param('fields', fields)

    def set_script_fields(self, script_fields):
        """
        Set script fields
        @param script_fields:
        @type script_fields: dict or pylastica.scriptfields.ScriptFields
        @return:
        @rtype: self
        """
        if isinstance(script_fields, dict):
            script_fields = pylastica.scriptfields.ScriptFields(script_fields)
        return self.set_param('script_fields', script_fields.to_dict())

    def add_script_field(self, name, script):
        """
        Add a script to the query
        @param name: name of the script
        @type name: str
        @param script:
        @type script: pylastica.script.Script
        @return:
        @rtype: self
        """
        assert isinstance(script, pylastica.script.Script), "script must be an instance of Script: %r" % script
        self._params['script_fields'][name] = script.to_dict()
        return self

    def set_facets(self, facets):
        """
        Sets all facets for this query object. Replaces existing facets
        @param facets:
        @type facets: dict of pylastica.facet.AbstractFacet
        @return:
        @rtype: self
        """
        self._params['facets'] = {}
        for facet in facets:
            self.add_facet(facet)
        return self

    def add_facet(self, facet):
        """
        Add a facet to the query
        @param facet:
        @type facet: pylastica.facet.AbstractFacet
        @return:
        @rtype: self
        """
        assert isinstance(facet, pylastica.facet.AbstractFacet), "facet must be an instance of an implementation of AbstractFacet: %r" % facet
        if 'facets' not in self._params:
            self._params['facets'] = {}
        self._params['facets'][facet.name] = facet.to_dict()
        return self

    def add_aggregation(self, aggregation):
        """
        Add an aggregation to the query
        @param aggregation:
        @type aggregation: pylastica.aggregation.AbstractAggregation
        @return:
        @rtype: self
        """
        if not isinstance(aggregation, AbstractAggregation):
            raise TypeError("aggregation must be an instance of an implementation of AbstractAggregation: %r" % aggregation)
        if 'aggs' not in self._params:
            self._params['aggs'] = {}
        self._params['aggs'][aggregation.name] = aggregation.to_dict()
        return self

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        #if no query is set, all query is chosen by default
        if 'query' not in self._params:
            self.query = pylastica.query.MatchAll()
        return self._params

    def set_min_score(self, min_score):
        """
        Allows filtering of documents based on a minimum score
        @param min_score:
        @type min_score: float
        @return:
        @rtype: self
        """
        assert isinstance(min_score, int) or isinstance(min_score, float), "min_score must be either a float or int: %r" % min_score
        return self.set_param('min_score', min_score)

    def set_suggest(self, suggest):
        """

        @param suggest:
        @type suggest: pylastica.suggest.Suggest
        @return:
        @rtype: self
        """
        self._params.update(suggest.to_dict())
        return self



__author__ = 'Joe Linn'

import pylastica.aggregation.abstractaggregation as abstract


class Histogram(abstract.SimpleAggregation):
    def __init__(self, name, field, interval):
        """
        @param name: the name of this aggregation
        @type name: str
        @param field: the field on which to perform this aggregation
        @type field: str
        @param interval: the interval by which documents will be bucketed
        @type interval: int
        """
        super(Histogram, self).__init__(name)
        self.set_field(field).set_interval(interval)

    def set_interval(self, interval):
        """
        Set the interval by which documents will be bucketed
        @param interval:
        @type interval: int
        @return:
        @rtype: Histogram
        """
        return self.set_param("interval", interval)

    def set_order(self, order, direction):
        """
        Set the bucket sort order
        @param order: _count, _term, or the name of a sub-aggregation or sub-aggregation response field
        @type order: str
        @param direction: asc or desc
        @type direction: str
        @return:
        @rtype: Histogram
        """
        return self.set_param('order', {order: direction})

    def set_minimum_document_count(self, count):
        """
        Set the minimum number of documents which must fall into a bucket in order for the bucket to be returned
        @param count: set to 0 to include empty buckets
        @type count: int
        @return:
        @rtype: Histogram
        """
        return self.set_param('min_doc_count', count)


__author__ = 'Joe Linn'

import pylastica.aggregation.abstractaggregation as abstract


class Missing(abstract.AbstractAggregation):
    def __init__(self, name, field):
        """
        @param name: the name of this aggregation
        @type name: str
        @param field: the name of the document field on which to perform this aggregation
        @type field: str
        """
        super(Missing, self).__init__(name)
        self.set_field(field)

    def set_field(self, field):
        """
        Set the field for this aggregation
        @param field: the name of the document field on which to perform this aggregation
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param("field", field)